
!! Copyright (C) 2004-2012 M. Oliveira, F. Nogueira, T. Cerqueira
!!
!! This program is free software; you can redistribute it and/or modify
!! it under the terms of the GNU General Public License as published by
!! the Free Software Foundation; either version 2, or (at your option)
!! any later version.
!!
!! This program is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU General Public License for more details.
!!
!! You should have received a copy of the GNU General Public License
!! along with this program; if not, write to the Free Software
!! Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
!! 02110-1301, USA.
!!
!! $Id: atom_of_inc.F90  2017-09-04 21:33  Kai Luo, University of Florida
!! KL modified 2017-09-25
!! KL modified 2017-11-16 include the real LPP of BLPS format in the pseudopotential, pspcod=8

  subroutine atom_create_of(atm, eigensolver)
    !-----------------------------------------------------------------------!
    ! Initializes the atom by reading the relevant input options and prints !
    ! some relevant informantion.                                           !
    !-----------------------------------------------------------------------!
    type(atom_t),        intent(inout) :: atm
    type(eigensolver_t), intent(in)    :: eigensolver
    integer :: ikedf ! index for different KEDFs, see subroutine pauli_potential
    ! only for GGA currently
    integer :: oftype
    ! oftype = 0 with all electron potential
    ! oftype = 1 with local pseudopotential

    real(R8), allocatable :: vh(:,:), vxc(:,:), vxctau(:,:),  vth(:,:), tth(:,:)
    real(R8), allocatable :: vloc(:), vtot(:)

    integer    :: unit
    integer    :: i,j
    integer    :: numline, a(5)
	  real(R8)   :: r
    real(R8), allocatable :: rtmp(:), vloctmp(:)
    !real(R8), allocatable :: vlpp(:)


    type(spline_t) :: vspline

    character(len=50) :: str

    !integer :: kedf
    call push_sub("atom_create_of")

                    !---Read information about the atom---!


    call oct_parse_int('ikedf', KEDF_NONE, ikedf)

    !Read theory level
    call oct_parse_int('TheoryLevel', DFT, atm%theory_level)
    select case (atm%theory_level)
    case (INDEPENDENT_PARTICLES, DFT)
    case default
      message(1) = "Illegal TheoryLevel."
      call write_fatal(1)
    end select

    !Read nuclear charge
    call oct_parse_float('NuclearCharge', M_ZERO, atm%z)
    ! if (atm%z <= M_ZERO) then
    !   message(1) = "Wrong or absent value for the nuclear charge."
    !   call write_fatal(1)
    ! end if
    atm%symbol = label(atm%z)

    !Read rel
    call  oct_parse_int('WaveEquation', SCHRODINGER, atm%wave_eq)
    select case (atm%wave_eq)
    case (SCHRODINGER, DIRAC, SCALAR_REL)
    case default
      message(1) = "Illegal WaveEquation."
      call write_fatal(1)
    end select

    !Read polarization
    call oct_parse_int('SpinMode', 1, atm%nspin)
    select case (atm%nspin)
    case (1,2)
    case default
      message(1) = "Illegal SpinMode."
      call write_fatal(1)
    end select


                    !---Initialize the data structures---!

    atm%type = ATOM_OF

    !Initialize xc
    call xc_null(atm%xc_model)
    if (atm%theory_level == DFT) call xc_init(atm%nspin, atm%xc_model)

    !Initialize mesh
    call mesh_null(atm%m)
    call mesh_init_from_input(atm%z, atm%m)

    !Initialize orbitals from file
    call atom_init_states_from_block(atm, "OFOrbitals")

    !Initialize potential
    allocate(vh(atm%m%np, atm%nspin), vxc(atm%m%np, atm%nspin), vxctau(atm%m%np, atm%nspin))

    !need allocate memory !!!
    allocate(vth(atm%m%np, atm%nspin))
    allocate(tth(atm%m%np, atm%nspin))

    allocate(vloc(atm%m%np), vtot(atm%m%np))

    select case (atm%theory_level)
    case (INDEPENDENT_PARTICLES)
      vh = M_ZERO; vxc = M_ZERO; vxctau = M_ZERO
    case (DFT)
      call hartree_potential(atm%m, states_batch_charge_density(atm%states, atm%m), &
                             states_batch_charge(atm%states), vh = vh)
      call xc_potential(atm%xc_model, atm%m, atm%states, atm%nspin, vxc = vxc, vxctau=vxctau)
    end select

    !
    !
    ! fixed the counter intuitive code, either all-electron or local pseudo potential
    ! read in the type of potential, either all-electron or local pseudopotential
    ! oftype code 0 ae
    ! oftype code 1 ps
    call oct_parse_int('oftype', 0, oftype)
    !write(*,*) "oftype= ", oftype

    call pauli_potential(ikedf, atm%m, atm%nspin, atm%states, vth, tth)
    if(oftype == 0) then
      call potential_init(atm%potential, atm%m, atm%nspin, vh + vxc + vth, vxctau, atm%z)
    else if (oftype == 1)  then
      call load_lpp(numline, rtmp, vloctmp)

!      do i=1, atm%m%np
!        r=atm%m%r(i)
!	    do j=1,numline-1
!            if ( (r-rtmp(j))*(r-rtmp(j+1)) <= 0) then
!                exit
!            end if
!        end do
!        vloc(i) = vloctmp(j)+(vloctmp(j+1)-vloctmp(j))/(rtmp(j+1)-rtmp(j))*(r-rtmp(j))
!      end do

      !write(*,*) "numline=", numline
      !call spline_mesh_transfer(numline, rtmp, vloctmp, atm%m%np, atm%m%r, vloc, 3)
      !call spline_init(vspline, numline, vlpp(:,1), vlpp(:,2), 3)

      !vtot=vloc+vth(:,1)
      call spline_null(vspline)
	    call spline_init(vspline, numline, rtmp, vloctmp,3)
      call io_open(unit, file="of/locpot.dat")
      do j=1, atm%m%np
        vloc(j)=spline_eval(vspline, atm%m%r(j))
        write(unit,*) atm%m%r(j), vloc(j)
      end do
      close(unit)

      ! this calls potential_init_ps_scr, which I do not need to redefine it for of-dft
      call potential_init(atm%potential, atm%m, atm%nspin, vh+vxc+vth, vxctau, vloc)

      !write(*,*)  "vth=",vth
      !
      ! need to load the local pseudo-potential here
      !call potential_init(atm%potential, atm%m, atm%nspin, vh + vxc, vxctau, atm%z)

      deallocate(rtmp, vloctmp)
    end if

    deallocate(vh, vxc, vxctau)
    deallocate(vtot, vth, vloc)

    !Initialize integrator
    call integrator_init(atm%integrator_sp, atm%integrator_dp)


                    !---Print information about the atom---!

    message(1) = ""
    message(2) = "General Information about the atom:"
    call write_info(2)
    call write_info(2,unit=info_unit("of"))

    !Print info to screen and to the of/info file
    write(message(1),'(2X,"Symbol: ",A)') trim(atm%symbol)
    select case (atm%theory_level)
    case (INDEPENDENT_PARTICLES)
      message(2) = "  Theory Level: independent particles"
    case (DFT)
      message(2) = "  Theory Level: DFT"
    end select
    select case (atm%wave_eq)
    case (SCHRODINGER)
      message(3) = "  Wave-equation: Schrodinger"
    case (SCALAR_REL)
      message(3) = "  Wave-equation: scalar-relativistic"
    case (DIRAC)
      message(3) = "  Wave-equation: Dirac"
    end select
    if (atm%nspin == 1) then
      message(4) = "  Spin mode: unpolarized"
    else
      message(4) = "  Spin mode: polarized"
    end if
    write(message(5),'(2X,"Nuclear charge: ",F6.2)') atm%z
    write(message(6),'(2X,"Total charge:   ",F6.2)') atm%z - states_batch_charge(atm%states)
    call write_info(6, verbose_limit=20)
    call write_info(6, unit=info_unit("of"))

    call states_batch_sort(atm%states, SORT_QN)
    call states_batch_output_configuration(atm%states, atm%nspin, verbose_limit=20)
    call states_batch_output_configuration(atm%states, atm%nspin, unit=info_unit("of"))

    if (atm%theory_level == DFT) then
      call xc_model_output_info(atm%xc_model, verbose_limit=20)
      call xc_model_output_info(atm%xc_model, unit=info_unit("of"))
    end if

    call mesh_output_params(atm%m, verbose_limit=20)
    call mesh_output_params(atm%m, unit=info_unit("of"))


                    !---Solve Khon-Sham equation---!

    call atom_solve(atm, eigensolver, info_unit("of"))

    call pop_sub()
  end subroutine atom_create_of


  subroutine load_lpp(mmax, r, vloc)

    !character(len=200), intent(in) :: fname
    real(R8), allocatable, intent(inout) :: r(:), vloc(:)


    character(len=200) :: title
    integer :: pspdat, pspcod, pspxc, lmax, lloc, pspso0
    integer, intent(out) :: mmax
    real(R8) :: zion,znucl, r2well
    real(R8) :: aa, bb
    real(R8) :: rchrg,fchrg,qchrg

    integer :: unit
    integer :: ii, jj, kk
    integer :: dummy


    call push_sub("load_lpp")
    call io_open(unit, file='lpp')

    read (unit,'(a)') title
    read (unit,*) znucl,zion,pspdat
    read (unit,*) pspcod,pspxc,lmax,lloc,mmax,r2well

    allocate(r(mmax),vloc(mmax))

    if( pspcod==8) then
        read(unit, *) rchrg,fchrg,qchrg
        read(unit, *) ! nproj
        read(unit, *) ! extension_switch
        read(unit, *) ! 0=l for pseudopotential
        do ii=1,mmax
            read(unit, *) dummy, r(ii), vloc(ii)
            !write(*,*) r(ii), vloc(ii)
        end do

    else if (pspcod == 5) then
        read (unit,*) aa,bb,pspso0


        do ii=1,mmax
        r(ii)=aa*exp(dble(ii-1)*bb)
        end do

        do jj=1,lmax+1
        read(unit, *)
        read(unit, *)
        end do

        read(unit, *) rchrg,fchrg,qchrg

        read(unit, *) ! 0=l for pseudopotential

        read(unit, *) (vloc(ii),ii=1,mmax)
    end if


    !do kk=1,mmax
    !  write(*,*) r(kk), vloc(kk)
    !end do
    ! discard all others
    close(unit)
    !deallocate(r,vloc)

    call pop_sub()

  end subroutine load_lpp
