!! Copyright (C) 2011-2012 M. Oliveira
!!
!! This program is free software; you can redistribute it and/or modify
!! it under the terms of the GNU General Public License as published by
!! the Free Software Foundation; either version 2, or (at your option)
!! any later version.
!!
!! This program is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU General Public License for more details.
!!
!! You should have received a copy of the GNU General Public License
!! along with this program; if not, write to the Free Software
!! Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
!! 02110-1301, USA.
!!
!! $Id: states_batch.F90 784 2013-08-02 15:21:03Z micael $

#include "global.h"

module states_batch_m
  use global_m
  use oct_parser_m
  use messages_m
  use units_m
  use io_m
  use output_m
  use quantum_numbers_m
  use mesh_m
  use ps_io_m
  use potentials_m
  use wave_equations_m
  use eigensolver_m
  use states_m

  implicit none


                    !---Interfaces---!

  interface assignment (=)
    module procedure states_batch_copy
  end interface


                    !---Derived Data Types---!

  type states_batch_t
    private
    integer :: n_states                   ! Number of states in the batch
    type(state_ptr), pointer :: states(:) ! Pointers to the states
  end type states_batch_t

  type state_ptr
    type(state_t), pointer :: ptr ! Pointer to the state
  end type state_ptr


                    !---Global Variables---!

  integer, parameter :: SORT_EV = 1, &
                        SORT_QN = 2

  integer, parameter :: OCC_FIXED           = 0, &
                        OCC_SEMICONDUCTING  = 1, &
                        OCC_AVERILL_PAINTER = 2


                    !---Public/Private Statements---!

  private
  public :: states_batch_t, &
            states_batch_null, &
            states_batch_init, &
            states_batch_end, &
            states_batch_deallocate, &
            assignment(=), &
            states_batch_size, &
            states_batch_add, &
            states_batch_get, &
            states_batch_number_of_folds, &
            states_batch_eigenvalues, &
            states_batch_split_folds, &
            states_batch_density, &
            states_batch_density_grad, &
            states_batch_density_lapl, &
            states_batch_charge_density, &
            states_batch_magnetization_density, &
            states_batch_tau, &
            states_batch_density_moment, &
            states_batch_charge, &
            states_batch_max_charge, &
            states_batch_ip, &
            states_batch_eigensolve, &
            states_batch_sort, &
            states_batch_smearing, &
            states_batch_psp_generation, &
            states_batch_ld_test, &
            states_batch_output_configuration, &
            states_batch_output_eigenvalues, &
            states_batch_output_density, &
            states_batch_output_shellstructure, &
            states_batch_ps_io_set, &
            !KL  2017-10-16
            pauli_potential, &
            output_pauli_potential, &

            OCC_FIXED, OCC_SEMICONDUCTING, OCC_AVERILL_PAINTER, &
            SORT_EV, SORT_QN


contains

  subroutine states_batch_null(batch)
    !-----------------------------------------------------------------------!
    !  Nullifies and sets to zero all the components of the batch.          !
    !-----------------------------------------------------------------------!
    type(states_batch_t), intent(out) :: batch

    call push_sub("states_batch_null")

    batch%n_states = 0
    nullify(batch%states)

    call pop_sub()
  end subroutine states_batch_null

  subroutine states_batch_init(batch, n_states, states)
    !-----------------------------------------------------------------------!
    ! Initializes a batch from an array of states.                          !
    !                                                                       !
    !  batch    - the batch to be initialized                               !
    !  n_states - the number of states                                      !
    !  states   - the array of states                                       !
    !-----------------------------------------------------------------------!
    type(states_batch_t),  intent(inout) :: batch
    integer,               intent(in)    :: n_states
    type(state_t), target, intent(in)    :: states(:)

    integer :: i

    call push_sub("states_batch_init")

    batch%n_states = n_states
    allocate(batch%states(n_states))
    do i = 1, n_states
      batch%states(i)%ptr => states(i)
    end do

    call pop_sub()
  end subroutine states_batch_init

  subroutine states_batch_end(batch)
    !-----------------------------------------------------------------------!
    ! Frees all memory associated to a batch                                !
    !-----------------------------------------------------------------------!
    type(states_batch_t), intent(inout) :: batch

    call push_sub("states_batch_end")

    if (associated(batch%states)) deallocate(batch%states)
    batch%n_states = 0

    call pop_sub()
  end subroutine states_batch_end

  subroutine states_batch_deallocate(batch)
    !-----------------------------------------------------------------------!
    ! Frees all memory associated to the states that are pointed to by the  !
    ! batch and then frees the memory associated to the batch object itself.!
    !-----------------------------------------------------------------------!
    type(states_batch_t), intent(inout) :: batch

    integer :: i

    call push_sub("states_batch_deallocate")

    if (associated(batch%states)) then
      do i = 1, batch%n_states
        call state_end(batch%states(i)%ptr)
      end do
      deallocate(batch%states)
    end if
    batch%n_states = 0

    call pop_sub()
  end subroutine states_batch_deallocate

  subroutine states_batch_copy(batch_a, batch_b)
    !-----------------------------------------------------------------------!
    ! Copy batch_b to batch_a                                               !
    !-----------------------------------------------------------------------!
    type(states_batch_t),  intent(inout) :: batch_a
    type(states_batch_t),  intent(in)    :: batch_b

    integer :: i

    call push_sub("states_batch_copy")

    call states_batch_end(batch_a)

    batch_a%n_states = batch_b%n_states
    allocate(batch_a%states(batch_a%n_states))
    do i = 1, batch_a%n_states
      batch_a%states(i)%ptr => batch_b%states(i)%ptr
    end do

    call pop_sub()
  end subroutine states_batch_copy

  function states_batch_size(batch)
    !-----------------------------------------------------------------------!
    ! Return the number of states in the batch.                             !
    !-----------------------------------------------------------------------!
    type(states_batch_t), intent(in) :: batch
    integer :: states_batch_size

    call push_sub("states_batch_size")

    states_batch_size = batch%n_states

    call pop_sub()
  end function states_batch_size

  subroutine states_batch_add(batch, state)
    !-----------------------------------------------------------------------!
    ! Add a state to a bacth.                                               !
    !                                                                       !
    !  batch - the batch                                                    !
    !  state - the state to be added to the batch                           !
    !-----------------------------------------------------------------------!
    type(states_batch_t),  intent(inout) :: batch
    type(state_t), target, intent(in)    :: state

    type(state_ptr), allocatable :: tmp(:)

    call push_sub("states_batch_add")

    if (batch%n_states == 0) then
      batch%n_states = 1
      allocate(batch%states(batch%n_states))
    else
      allocate(tmp(batch%n_states))
      tmp = batch%states
      deallocate(batch%states)
      allocate(batch%states(batch%n_states + 1))
      batch%states(1:batch%n_states) = tmp(1:batch%n_states)
      batch%n_states = batch%n_states + 1
    end if
    batch%states(batch%n_states)%ptr => state

    call pop_sub()
  end subroutine states_batch_add

  function states_batch_get(batch, i)
    !-----------------------------------------------------------------------!
    ! Returns a pointer to the ith state of the batch.                      !
    !-----------------------------------------------------------------------!
    type(states_batch_t), intent(in) :: batch
    integer,              intent(in) :: i
    type(state_t), pointer :: states_batch_get

    call push_sub("states_batch_get")

    ASSERT(i > 0 .and. i <= batch%n_states)
    states_batch_get => batch%states(i)%ptr

    call pop_sub()
  end function states_batch_get

  function states_batch_eigenvalues(batch)
    !-----------------------------------------------------------------------!
    ! Returns a pointer to the ith state of the batch.                      !
    !-----------------------------------------------------------------------!
    type(states_batch_t), intent(in) :: batch
    real(R8) :: states_batch_eigenvalues(batch%n_states)

    integer :: i

    call push_sub("states_batch_eigenvalues")

    do i = 1, batch%n_states
      states_batch_eigenvalues(i) = state_eigenvalue(batch%states(i)%ptr)
    end do

    call pop_sub()
  end function states_batch_eigenvalues

  function states_batch_number_of_folds(batch, polarized)
    !-----------------------------------------------------------------------!
    ! Given a batch, return the number of quantum number folds that exist   !
    ! in the bacth (see quantum_numbers_m for the definition of a fold).    !
    !-----------------------------------------------------------------------!
    type(states_batch_t), intent(in) :: batch
    logical,              intent(in) :: polarized
    integer :: states_batch_number_of_folds

    integer :: i
    type(qn_t) :: qn
    logical, allocatable :: checked(:)

    call push_sub("states_batch_number_of_folds")

    states_batch_number_of_folds = 0

    if (batch%n_states == 0) return

    allocate(checked(batch%n_states))
    checked = .false.
    do
      if (all(checked)) exit

      do i = 1, batch%n_states
        if (.not. checked(i)) then
          states_batch_number_of_folds = states_batch_number_of_folds + 1
          qn = state_qn(batch%states(i)%ptr)
          exit
        end if
      end do

      do i = 1, batch%n_states
        if (.not. checked(i) .and. qn_equal_fold(qn, state_qn(batch%states(i)%ptr), polarized)) then
          checked(i) = .true.
        end if
      end do
    end do

    ASSERT(all(checked))
    deallocate(checked)

    call pop_sub()
  end function states_batch_number_of_folds

  function states_batch_density(batch, nspin, m)
    !-----------------------------------------------------------------------!
    ! Computes the electronic density associated with a batch of states.    !
    !-----------------------------------------------------------------------!
    type(states_batch_t), intent(in) :: batch
    integer,              intent(in) :: nspin
    type(mesh_t),         intent(in) :: m
    real(R8) :: states_batch_density(m%np, nspin)

    integer :: i

    call push_sub("states_batch_density")

    states_batch_density = M_ZERO
    do i = 1, batch%n_states
      states_batch_density = states_batch_density + &
                             state_density(nspin, batch%states(i)%ptr)
    end do

    call pop_sub()
  end function states_batch_density

  function states_batch_density_grad(batch, nspin, m)
    !-----------------------------------------------------------------------!
    ! Computes the gradient of the electronic density associated with a     !
    ! batch of states.                                                      !
    !-----------------------------------------------------------------------!
    type(states_batch_t), intent(in) :: batch
    integer,              intent(in) :: nspin
    type(mesh_t),         intent(in) :: m
    real(R8) :: states_batch_density_grad(m%np, nspin)

    integer :: i

    call push_sub("states_batch_density_grad")

    states_batch_density_grad = M_ZERO
    do i = 1, batch%n_states
      states_batch_density_grad = states_batch_density_grad + &
                               state_density_grad(nspin, batch%states(i)%ptr, m)
    end do

    call pop_sub()
  end function states_batch_density_grad

  function states_batch_density_lapl(batch, nspin, m)
    !-----------------------------------------------------------------------!
    ! Computes the laplacian of the electronic density associated with a    !
    ! batch of states.                                                      !
    !-----------------------------------------------------------------------!
    type(states_batch_t), intent(in) :: batch
    integer,              intent(in) :: nspin
    type(mesh_t),         intent(in) :: m
    real(R8) :: states_batch_density_lapl(m%np, nspin)

    integer :: i

    call push_sub("states_batch_density_lapl")

    states_batch_density_lapl = M_ZERO
    do i = 1, batch%n_states
      states_batch_density_lapl = states_batch_density_lapl + &
                               state_density_lapl(nspin, batch%states(i)%ptr, m)
    end do

    call pop_sub()
  end function states_batch_density_lapl

  function states_batch_charge_density(batch, m)
    !-----------------------------------------------------------------------!
    ! Computes the charge density associated with a batch of states.        !
    !-----------------------------------------------------------------------!
    type(states_batch_t), intent(in) :: batch
    type(mesh_t),         intent(in) :: m
    real(R8) :: states_batch_charge_density(m%np)

    integer :: i

    call push_sub("states_batch_charge_density")

    states_batch_charge_density = M_ZERO
    do i = 1, batch%n_states
      states_batch_charge_density = states_batch_charge_density + &
                             state_charge_density(batch%states(i)%ptr)
    end do

    call pop_sub()
  end function states_batch_charge_density

  function states_batch_magnetization_density(batch, m)
    !-----------------------------------------------------------------------!
    ! Computes the magnetization density associated with a batch of states. !
    !-----------------------------------------------------------------------!
    type(states_batch_t), intent(in) :: batch
    type(mesh_t),         intent(in) :: m
    real(R8) :: states_batch_magnetization_density(m%np)

    integer :: i

    call push_sub("states_batch_magnetization_density")

    states_batch_magnetization_density = M_ZERO
    do i = 1, batch%n_states
      states_batch_magnetization_density = states_batch_magnetization_density + &
                             state_magnetization_density(batch%states(i)%ptr)
    end do

    call pop_sub()
  end function states_batch_magnetization_density

  function states_batch_tau(batch, nspin, m)
    !-----------------------------------------------------------------------!
    ! Computes the kinetic energy density associated with a batch of states.!
    !-----------------------------------------------------------------------!
    type(states_batch_t), intent(in) :: batch
    integer,              intent(in) :: nspin
    type(mesh_t),         intent(in) :: m
    real(R8) :: states_batch_tau(m%np, nspin)

    integer :: i

    call push_sub("states_batch_tau")

    states_batch_tau = M_ZERO
    do i = 1, batch%n_states
      states_batch_tau = states_batch_tau + &
                               state_tau(nspin, batch%states(i)%ptr, m)
    end do

    call pop_sub()
  end function states_batch_tau

  function states_batch_density_moment(batch, m, order)
    !-----------------------------------------------------------------------!
    ! Computes the density moment sum_i < R_i(r) | r**order | R_i(r)> of a  !
    ! given batch of states.                                                !
    !-----------------------------------------------------------------------!
    type(states_batch_t), intent(in) :: batch
    type(mesh_t),         intent(in) :: m
    integer,              intent(in) :: order
    real(R8) :: states_batch_density_moment

    integer :: i

    call push_sub("states_batch_density_moment")

    states_batch_density_moment = M_ZERO
    do i = 1, batch%n_states
      states_batch_density_moment = states_batch_density_moment + &
                             state_density_moment(batch%states(i)%ptr, m, order)
    end do

  end function states_batch_density_moment

  subroutine states_batch_split_folds(batch, folds, polarized)
    !-----------------------------------------------------------------------!
    ! Given a batch, splits the states into folds (see quantum_numbers_m    !
    ! for the definition of a fold).                                        !
    !-----------------------------------------------------------------------!
    type(states_batch_t), intent(in)    :: batch
    type(states_batch_t), intent(inout) :: folds(:)
    logical,              intent(in)    :: polarized

    integer :: i, n
    type(qn_t) :: qn
    logical, allocatable :: checked(:)

    call push_sub("states_batch_split_folds")

    ASSERT(batch%n_states > 0)

    n = 0
    allocate(checked(batch%n_states))
    checked = .false.
    do
      if (all(checked)) exit
      n = n + 1
      call states_batch_end(folds(n))

      do i = 1, batch%n_states
        if (.not. checked(i)) then
          qn = state_qn(batch%states(i)%ptr)
          exit
        end if
      end do

      do i = 1, batch%n_states
        if (.not. checked(i) .and. qn_equal_fold(qn, state_qn(batch%states(i)%ptr), polarized)) then
          checked(i) = .true.
          call states_batch_add(folds(n), batch%states(i)%ptr)
        end if
      end do
    end do

    ASSERT(all(checked))
    deallocate(checked)

    call pop_sub()
  end subroutine states_batch_split_folds

  function states_batch_charge(batch)
    !-----------------------------------------------------------------------!
    ! Returns the number of electons in a batch.                            !
    !-----------------------------------------------------------------------!
    type(states_batch_t), intent(in) :: batch
    real(R8) :: states_batch_charge

    integer :: i

    call push_sub("states_batch_charge")

    states_batch_charge = M_ZERO
    do i = 1, batch%n_states
      states_batch_charge = states_batch_charge + state_charge(batch%states(i)%ptr)
    end do

    call pop_sub()
  end function states_batch_charge

  function states_batch_max_charge(batch)
    !-----------------------------------------------------------------------!
    ! Returns the maximum number of electons the batch can hold.            !
    !-----------------------------------------------------------------------!
    type(states_batch_t), intent(in) :: batch
    real(R8) :: states_batch_max_charge

    integer :: i

    call push_sub("states_batch_max_charge")

    states_batch_max_charge = M_ZERO
    do i = 1, batch%n_states
      states_batch_max_charge = states_batch_max_charge + qn_max_occ(state_qn(batch%states(i)%ptr))
    end do

    call pop_sub()

  end function states_batch_max_charge

  function states_batch_ip(batch, nspin)
    !-----------------------------------------------------------------------!
    ! Returns the ionization potential of a batch (defined as minus the     !
    ! eigenvalue of the highest occupied orbital) for each spin-channel.    !
    !-----------------------------------------------------------------------!
    type(states_batch_t), intent(in) :: batch
    integer,              intent(in) :: nspin
    real(R8) :: states_batch_ip(nspin)

    integer :: i, is
    real(R8) :: ev, charge
    type(qn_t) :: qn

    call push_sub("states_batch_ip")

    !Get ionization potential
    states_batch_ip = M_ZERO
    do i = 1, batch%n_states
      qn = state_qn(batch%states(i)%ptr)
      ev = state_eigenvalue(batch%states(i)%ptr)
      charge = state_charge(batch%states(i)%ptr)

      is = 1
      if (qn%s == M_HALF) is = 2

      if (charge /= M_ZERO) then
        if (states_batch_ip(is) == M_ZERO) then
          states_batch_ip(is) =  - ev
        elseif (-states_batch_ip(is) < ev) then
          states_batch_ip(is) = - ev
        end if
      end if
    end do

    call pop_sub()
  end function states_batch_ip

  subroutine states_batch_eigensolve(batch, m, wave_eq, potential, integrator_dp, integrator_sp, eigensolver)
    !-----------------------------------------------------------------------!
    ! Get the eigenvalues and eigenfunctions of a batch of states for a     !
    ! given potential.                                                      !
    !                                                                       !
    !  batch         - bacth of states                                      !
    !  m             - mesh                                                 !
    !  wave_eq       - wave-equation to solve                               !
    !  potential     - potential to use in the wave-equation                !
    !  integrator_sp - single-precision integrator object                   !
    !  integrator_dp - double-precision integrator object                   !
    !  eigensolver   - information about the eigensolver                    !
    !-----------------------------------------------------------------------!
    type(states_batch_t), intent(inout) :: batch
    type(mesh_t),         intent(in)    :: m
    integer,              intent(in)    :: wave_eq
    type(potential_t),    intent(in)    :: potential
    type(integrator_t),   intent(inout) :: integrator_dp
    type(integrator_t),   intent(inout) :: integrator_sp
    type(eigensolver_t),  intent(in)    :: eigensolver

    integer :: i, max_n_bound
    real(R8) :: ev
    type(qn_t) :: qn
    type(states_batch_t) :: bound
    logical, allocatable :: bracketed(:)
    real(R8), allocatable :: brackets(:,:)
    type(qn_t), allocatable :: qns(:)

    call push_sub("states_batch_eigensolve")

    !Batch should be ordered by increasing quantum numbers
    call states_batch_sort(batch, SORT_QN)

    !Check the maximum number of bound states we can have
    qn = state_qn(batch%states(1)%ptr)
    max_n_bound = wavefunctions_n_bound_states(qn, wave_eq, m, &
                                           potential, integrator_sp)
    !Remove states that we are sure are unbound from the list
    call states_batch_null(bound)
    do i = 1, batch%n_states
      qn = state_qn(batch%states(i)%ptr)

      if (qn%n - qn%l <= max_n_bound) then
        call states_batch_add(bound, batch%states(i)%ptr)
      else
        if (state_charge(batch%states(i)%ptr) == M_ZERO) then
          !Unbound state
          call state_update(batch%states(i)%ptr, m, wave_eq, potential, integrator_dp, ev=M_ZERO)
        else
          !There is something wrong
          write(message(1),'("State: ",A," is unbound")') &
               trim(state_label(batch%states(i)%ptr, .true.))
          !call write_fatal(1)
        end if
      end if
    end do

    !Bracket the eigenvalues
    allocate(brackets(2, bound%n_states), qns(bound%n_states), bracketed(bound%n_states))
    do i = 1, bound%n_states
      qns(i) = state_qn(bound%states(i)%ptr)
    end do
    call eigensolver_bracket(bound%n_states, qns, wave_eq, eigensolver, &
                             potential, integrator_sp, brackets, bracketed)

    !Find eigenvalues
    do i = 1, bound%n_states
      if (bracketed(i)) then
        call eigensolver_find_ev(qns(i), wave_eq, eigensolver, potential, &
                                 integrator_dp, brackets(:,i), ev)
      else if (.not. bracketed(i) .and. state_charge(bound%states(i)%ptr) == M_ZERO) then
        !This is probably and unbound state after all
        ev = M_ZERO
      else if (.not. bracketed(i) .and. state_charge(bound%states(i)%ptr) /= M_ZERO) then
        !There is something wrong
        message(1) = "Unable to bracket eigenvalues for state:"
        write(message(2),'(2X,A)') trim(qn_label(qns(i), .true.))
        call write_fatal(2)
      end if


      !Update the state with the new eigenvalue
      call state_update(bound%states(i)%ptr, m, wave_eq, potential, integrator_dp, ev)
    end do

    !Free memory
    deallocate(brackets, bracketed, qns)
    call states_batch_end(bound)

    call pop_sub()
  end subroutine states_batch_eigensolve

  subroutine states_batch_sort(batch, criteria)
    !-----------------------------------------------------------------------!
    ! Sort the states in the bacth.                                         !
    !                                                                       !
    !  batch    - bacth of states                                           !
    !  criteria - criteria used to sort the states                          !
    !-----------------------------------------------------------------------!
    type(states_batch_t), intent(inout) :: batch
    integer,              intent(in)    :: criteria

    integer :: i, dum(1)
    logical, allocatable :: mask(:), qn_mask(:)
    integer, allocatable :: order(:)
    real(R8), allocatable :: ev(:)
    type(qn_t), allocatable :: qn(:)
    type(state_ptr), allocatable :: tmp(:)

    call push_sub("states_batch_sort")

    allocate(mask(batch%n_states), order(batch%n_states), tmp(batch%n_states))

    !sort states
    mask = .true.
    select case (criteria)
    case (SORT_EV)
      allocate(ev(batch%n_states))
      do i = 1, batch%n_states
        ev(i) = state_eigenvalue(batch%states(i)%ptr)
      end do

      do i = 1, batch%n_states
        dum = maxloc(abs(ev), mask=mask)
        order(i) = dum(1)
        mask(order(i)) = .false.
      end do
      deallocate(ev)

    case (SORT_QN)
      !Quantum numbers by increasing n,l,j,m,s and decreasing sg
      !(the algorithm is not very efficient, but we should never have to deal
      ! with a very large number of states)
      allocate(qn(batch%n_states))
      do i = 1, batch%n_states
        qn(i) = state_qn(batch%states(i)%ptr)
      end do

      allocate(qn_mask(batch%n_states))
      do i = 1, batch%n_states
        qn_mask = qn%n == minval(qn%n, mask)
        qn_mask = qn_mask .and. qn%l == minval(qn%l, mask .and. qn_mask)
        qn_mask = qn_mask .and. qn%j == minval(qn%j, mask .and. qn_mask)
        qn_mask = qn_mask .and. qn%m == minval(qn%m, mask .and. qn_mask)
        qn_mask = qn_mask .and. qn%s == minval(qn%s, mask .and. qn_mask)
        dum = maxloc(qn%sg, mask = mask .and. qn_mask)
        order(i) = dum(1)
        mask(order(i)) = .false.
      end do
      deallocate(qn, qn_mask)

    end select

    tmp = batch%states
    do i = 1, batch%n_states
      batch%states(i) = tmp(order(i))
    end do

    deallocate(mask, order, tmp)

    call pop_sub()
  end subroutine states_batch_sort

  subroutine states_batch_smearing(batch, smearing_function, tol, new_charge)
    !-----------------------------------------------------------------------!
    ! Distributes the electrons among the states.                           !
    !                                                                       !
    !  smearing function - how should the electrons be distributed          !
    !  batch             - bacth of states                                  !
    !-----------------------------------------------------------------------!
    type(states_batch_t), intent(inout) :: batch
    integer,              intent(in)    :: smearing_function
    real(R8),             intent(in)    :: tol
    real(R8), optional,   intent(in)    :: new_charge

    integer :: n_states, is, i, qi, qf, k
    real(R8) :: total_charge, lambda, lambda_in, sum_evdiff
    integer,  allocatable :: order(:)
    real(R8), allocatable :: ev(:), occ(:), new_occ(:), max_occ(:), gamma(:)

    call push_sub("states_batch_smearing")

    if (smearing_function == OCC_FIXED) then
      ! Nothing to do
      call pop_sub()
      return
    end if

    n_states = batch%n_states

    !Sort states by increasing eigenvalue
    call states_batch_sort(batch, SORT_EV)

    !Allocate working arrays
    allocate(ev(batch%n_states), order(n_states))
    allocate(occ(n_states), max_occ(n_states), new_occ(n_states))

    !eigenvalues, occupancies, and maximum occupancies for all the states in a single vector
    do i = 1, n_states
      ev(i) = state_eigenvalue(batch%states(i)%ptr)
      occ(i) = state_charge(batch%states(i)%ptr)
      max_occ(i) = qn_max_occ(state_qn(batch%states(i)%ptr))
      order(i) = i
    end do

    !Order all arrays by decreasing occupancy
    if (n_states /= 1) then
      i = 1
      do
        i = i + 1
        if (ev(i-1) == ev(i) .and. occ(order(i-1)) < occ(order(i))) then
          is = order(i)
          order(i) = order(i-1)
          order(i-1) = is
          i = 1
        end if
        if (i == n_states) exit
      end do
    end if

    do i = 1, n_states
      new_occ(i) = occ(order(i))
    end do
    occ = new_occ
    do i = 1, n_states
      new_occ(i) = max_occ(order(i))
    end do
    max_occ = new_occ

    !
    select case (smearing_function)
    case (OCC_SEMICONDUCTING)

      if (present(new_charge)) then
        total_charge = new_charge
      else
        total_charge = states_batch_charge(batch)
      end if

      do i = 1, n_states
        if (total_charge == M_ZERO) then
          occ(i) = M_ZERO
        elseif (total_charge - max_occ(i) < M_ZERO) then
          occ(i) = total_charge
        else
          occ(i) = max_occ(i)
        end if
        total_charge = total_charge - occ(i)
      end do

    case (OCC_AVERILL_PAINTER)

      !Allocate memory and initialize some quantities
      allocate(gamma(n_states))
      lambda_in = M_HALF
      gamma = M_ONE
      total_charge = states_batch_charge(batch)

      !Remove states that will not participate in the charge transfer
      do i = 1, n_states
        if (occ(i) /= max_occ(i)) then
          qi = i
          exit
        end if
      end do
      qf = qi
      do i = n_states, qi, -1
        if (occ(i) /= M_ZERO) then
          qf = i
          exit
        end if
      end do

      !Main loop
      do
        !Exit if we have less than two states to perform the charge transfer
        if (count(gamma(qi:qf) /= M_ZERO) < 2) exit

        !By default we dont change the occupancies
        new_occ = occ

        !Choose the reference state
        if (occ(qi)/max_occ(qi) < M_HALF) then
          k = qi
        else
          k = qf
        end if

        !Compute gamma
        sum_evdiff = M_ZERO
        do i = qi, qf
          if (i == k) cycle
          if (abs(ev(i) - ev(k)) < tol) then
            gamma(i) = M_ZERO
          else if (ev(i) - ev(k) > M_ZERO) then
            gamma(i) = occ(i)/max_occ(i)/(ev(i) - ev(k))
          elseif (ev(i) - ev(k) < M_ZERO) then
            gamma(i) = (occ(i)/max_occ(i) - M_ONE)/(ev(i) - ev(k))
          end if
          if (gamma(i) /= M_ZERO) sum_evdiff = sum_evdiff + (ev(i) - ev(k))*max_occ(i)
        end do

        if (sum_evdiff > M_ZERO) then
          gamma(k) = (M_ONE - occ(k)/max_occ(k))/sum_evdiff
        elseif (sum_evdiff < M_ZERO) then
          gamma(k) = -occ(k)/max_occ(k)/sum_evdiff
        else
          gamma(k) = M_ZERO
        end if

        !If gamma(k) is zero remove this state from the list and go back to the
        !beginning of the cycle
        if (abs(gamma(k)) == M_ZERO) then
          if (k == qi) then
            qi = qi + 1
          else
            qf = qf - 1
          end if
          cycle
        end if

        !Get lambda (ignore the states that have gamma = 0)
        lambda = min(lambda_in, minval(gamma(qi:qf), mask=gamma(qi:qf) /= M_ZERO))

        !Get the new occupancies (states with gamma = 0 are ignored)
        sum_evdiff = M_ZERO
        do i = qi, qf
          if (i /= k .and. gamma(i) /= M_ZERO) then

            new_occ(i) = occ(i) - lambda*(ev(i) - ev(k))*max_occ(i)

            if (new_occ(i) < M_EPSILON) then
              new_occ(i) = M_ZERO
              sum_evdiff = sum_evdiff + occ(i)/lambda

            else if ( abs(max_occ(i) - new_occ(i)) < M_EPSILON ) then
              new_occ(i) = max_occ(i)
              sum_evdiff = sum_evdiff + (occ(i) - max_occ(i))/lambda

            else
              sum_evdiff = sum_evdiff + (ev(i) - ev(k))*max_occ(i)

            end if

          end if
        end do

        new_occ(k) = occ(k) + lambda*sum_evdiff
        if (new_occ(k) < M_EPSILON) then
          new_occ(k) = M_ZERO
        else if (max_occ(k) - new_occ(k) < M_EPSILON) then
          new_occ(k) = max_occ(k)
        end if

        occ = new_occ

        !Check if we have reached convergence
        if (lambda == lambda_in) exit

        lambda_in = lambda
      end do

      deallocate(gamma)
    end select

    !Get the final occupancies in the correct order
    do i = 1, n_states
      batch%states(order(i))%ptr%occ = occ(i)
    end do

    !Deallocate memory
    deallocate(ev, order, occ, max_occ, new_occ)

    call pop_sub()
  end subroutine states_batch_smearing

  subroutine states_batch_psp_generation(batch, m, scheme, wave_eq, tol, ae_potential, &
       integrator_sp, integrator_dp, eigensolver, rc, ps_v)
    !-----------------------------------------------------------------------!
    !-----------------------------------------------------------------------!
    type(states_batch_t), intent(inout) :: batch
    type(mesh_t),         intent(in)    :: m
    integer,              intent(in)    :: scheme, wave_eq
    real(R8),             intent(in)    :: tol
    type(potential_t),    intent(in)    :: ae_potential
    type(integrator_t),   intent(inout) :: integrator_sp, integrator_dp
    type(eigensolver_t),  intent(in)    :: eigensolver
    real(R8),             intent(in)    :: rc
    real(R8),             intent(out)   :: ps_v(m%np)

    integer :: n, i
    type(state_t), pointer :: state, state2
    type(potential_t) :: ps_potential
    type(states_batch_t) :: other_states

    call push_sub("states_batch_psp_generation")

    state => states_batch_get(batch, 1)
    if (scheme == MRPP .or. scheme == RMRPP) then
      n = 2
      state2 => states_batch_get(batch, 2)
      call state_psp_generation(m, scheme, wave_eq, tol, ae_potential, &
           integrator_sp, integrator_dp, ps_v, state, rc, state2)
    else
      n = 1
      call state_psp_generation(m, scheme, wave_eq, tol, ae_potential, &
           integrator_sp, integrator_dp, ps_v, state, rc)
    end if


    !We need to solve the wave-equation with the pseudopotential for the other states in the batch
    if (n < states_batch_size(batch)) then
      call states_batch_null(other_states)
      call potential_null(ps_potential)
      call potential_init(ps_potential, m, ps_v)

      do i = n+1, states_batch_size(batch)
        call states_batch_add(other_states, states_batch_get(batch, i))
      end do

      call states_batch_eigensolve(other_states, m, wave_eq, ps_potential, integrator_dp, integrator_sp, eigensolver)

      call potential_end(ps_potential)
      call states_batch_end(other_states)
    end if

    call pop_sub()
  end subroutine states_batch_psp_generation

  subroutine states_batch_ld_test(batch, m, ae_potential, ps_potential, integrator, r, de, emin, emax)
    !-----------------------------------------------------------------------!
    ! Computes the logarithmic derivatives for a given fold of states.      !
    !                                                                       !
    !  batch        - batch of states (they should belong to the same fold) !
    !  m            - mesh                                                  !
    !  ae_potential - all-electron potential                                !
    !  ps_potential - pseudotential                                         !
    !  integrator   - integrator object                                     !
    !  r            - diagnostic radius at which to do the test             !
    !  de           - energy step                                           !
    !  emin         - lower bound of the energy interval                    !
    !  emax         - upper bound of the energy intervel                    !
    !-----------------------------------------------------------------------!
    type(states_batch_t), intent(in) :: batch
    type(mesh_t),         intent(in) :: m
    type(potential_t),    intent(in) :: ae_potential
    type(potential_t),    intent(in) :: ps_potential
    type(integrator_t),   intent(inout) :: integrator
    real(R8),             intent(in) :: r
    real(R8),             intent(in) :: de
    real(R8), optional,   intent(in) :: emin, emax

    integer :: unit
    real(R8) :: e, e1, e2, ae_ld, ps_ld, step, dldde
    character(len=10) :: label

    call push_sub("states_batch_ld_test")

    ASSERT(batch%n_states > 0)

    !Set energy interval
    if (present(emin)) then
      e1 = emin
    else
      e1 = minval(states_batch_eigenvalues(batch)) - M_ONE
    end if
    if (present(emax)) then
      e2 = emax
    else
      e2 = maxval(states_batch_eigenvalues(batch)) + M_ONE
    end if
    if (e1 > e2) then
      message(1) = "The minimum energy at which to calculate the logarithmic"
      message(2) = "derivatives must be smaller than the maximum energy."
      call write_fatal(2)
    end if

    !Open file
    label = state_label(batch%states(1)%ptr)
    call io_open(unit, 'tests/ld-'//trim(label(2:5)))

    !Write some information
    write(message(1), '(2X,"Computing logarithmic derivative for states:",1X,A)') &
         trim(label(2:5))
    write(message(2), '(4X,"Minimum energy: ",F6.3,1X,A)') &
         e1/units_out%energy%factor, trim(units_out%energy%abbrev)
    write(message(3), '(4X,"Maximum energy: ",F6.3,1X,A)') &
         e2/units_out%energy%factor, trim(units_out%energy%abbrev)
    call write_info(3)
    call write_info(3,unit=info_unit("tests"))

    !Write header
    write(unit,'("# Logarithmic derivatives")')
    write(unit,'("# ")')
    write(unit,'("# Energy units: ",A)') trim(units_out%energy%name)
    write(unit,'("# Length units: ",A)') trim(units_out%length%name)
    write(unit,'("#")')
    write(unit,'("# ",53("-"))')
    write(unit,'("# |",7X,"e",7X,"|",5X,"ld_ae(e)",4X,"|",5X,"ld_pp(e)",4X,"|")')
    write(unit,'("# ",53("-"))')

    !Compute the logarithmic derivatives
    e = e1
    do
      ae_ld = state_ld(batch%states(1)%ptr, e, r, integrator, ae_potential, m)
      if (de == M_ZERO) then
        ps_ld = state_ld(batch%states(1)%ptr, e, r, integrator, ps_potential, m, dldde)
      else
        ps_ld = state_ld(batch%states(1)%ptr, e, r, integrator, ps_potential, m)
      end if

      write(unit,'(3X,ES14.7E2,3X,ES15.8E2,3X,ES15.8E2)') e/units_out%energy%factor, ae_ld, ps_ld

      if (e == e2) exit
      if (de /= M_ZERO) then
        step = de
      else
        step = max(abs(0.1_r8/dldde), 0.01_r8)
      end if
      e = min(e + step, e2)
    end do

    close(unit)

    call pop_sub()
  end subroutine states_batch_ld_test

  subroutine states_batch_output_configuration(batch, nspin, unit, verbose_limit)
    !-----------------------------------------------------------------------!
    ! Outputs the states label and occupancies in a nice readable format.   !
    !-----------------------------------------------------------------------!
    type(states_batch_t), intent(in) :: batch
    integer,              intent(in) :: nspin
    integer,              intent(in), optional :: unit, verbose_limit

    integer :: i
    character(len=4) :: nspin_label
    type(qn_t) :: qn
    type(state_t), pointer :: state

    call push_sub("states_batch_output_configuration")

    select case (nspin)
    case (1)
      write(message(1),'(2X,"Configuration    : State   Occupation")')
    case (2)
      qn = state_qn(batch%states(1)%ptr)
      if (qn%m == M_ZERO) then
        write(message(1),'(2X,"Configuration    : State   Spin   Occupation")')
      else
        write(message(1),'(2X,"Configuration    :  State   Sigma   Occupation")')
      end if
    end select

    do i =1, batch%n_states
      state => batch%states(i)%ptr
      qn = state_qn(state)
      select case (nspin)
      case (1)
        if (qn%j == M_ZERO) then
          write(message(i+1),'(23X,A,6X,F5.2)') trim(state_label(state)), state_charge(state)
        else
          write(message(i+1),'(21X,A,5X,F5.2)') trim(state_label(state)), state_charge(state)
        end if
      case (2)
        if (qn%m == M_ZERO) then
          if (qn%s == -M_HALF) then
            nspin_label = "-1/2"
          else
            nspin_label = "+1/2"
          end if
          write(message(i+1),'(21X,A,4X,A,5X,F5.2)') trim(state_label(state)), &
                                                     nspin_label, state_charge(state)
        else
          if (qn%sg == M_HALF) then
            nspin_label = 'up'
          elseif (qn%sg == -M_HALF) then
            nspin_label = 'dn'
          else
            nspin_label = '--'
          end if
          write(message(i+1),'(21X,A,5X,A,4X,F5.2)') trim(state_label(state)), &
                                                     nspin_label, state_charge(state)
        end if
      end select
    end do

    if (present(unit)) then
      call write_info(batch%n_states+1, unit=unit)
    else
      if (present(verbose_limit)) then
        call write_info(batch%n_states+1, verbose_limit)
      else
        call write_info(batch%n_states+1)
      end if
    end if

    call pop_sub()
  end subroutine states_batch_output_configuration

  subroutine states_batch_output_eigenvalues(batch, nspin, unit, verbose_limit)
    !-----------------------------------------------------------------------!
    ! Writes the eigenvalues either to a file or to the screen in a nice    !
    ! readable format.                                                      !
    !                                                                       !
    !  batch - batch of states                                              !
    !-----------------------------------------------------------------------!
    type(states_batch_t), intent(in) :: batch
    integer,              intent(in) :: nspin
    integer,              intent(in), optional :: unit, verbose_limit

    integer  :: i, j
    real(R8) :: u
    type(qn_t) :: qn
    character(len=4) :: nspin_label
    type(state_t), pointer :: state

    call push_sub("states_batch_output_eigenvalues")

    ! KL 07/25/2017 output more digits of the eigenvalues
    !

    !write(message(1),'(4X,"Eigenvalues [",A,"]")') trim(units_out%energy%abbrev)
    write(message(1),'("Eigenvalues [",A,"]")') trim(units_out%energy%abbrev)

    select case (nspin)
    case (1)
      !write(message(2),'(5X,"State   Occupation   Eigenvalue")')
      write(message(2),'(7X,"State", 5X,"Occupation", 5X, "Eigenvalue")')
    case (2)
      qn = state_qn(batch%states(1)%ptr)
      if (qn%m == M_ZERO) then
        !write(message(2),'(5X,"State   Spin   Occupation   Eigenvalue")')
        write(message(2),'(7X,"State", 5X, "Sigma", 5X,"Occupation", 5X, "Eigenvalue")')
      else
        !write(message(2),'(5X," State   Sigma   Occupation   Eigenvalue")')
        write(message(2),'(7X,"State", 5X, "Sigma", 5X,"Occupation", 5X, "Eigenvalue")')
      end if
    end select

    u = units_out%energy%factor
    j = 2
    do i = 1, batch%n_states
      j = j + 1
      state => batch%states(i)%ptr

      qn = state_qn(state)
      select case (nspin)
      case (1)
        if (qn%j == M_ZERO) then
          !write(message(j),'(7X,A,6X,F5.2,4X,F12.5)') trim(state_label(state)), &
          write(message(j),'(7X,A,6X,F5.2,6X,F15.8)') trim(state_label(state)), &
                                state_charge(state), state_eigenvalue(state)/u
        else
          !write(message(j),'(5X,A,5X,F5.2,4X,F12.5)') trim(state_label(state)), &
          write(message(j),'(5X,A,5X,F5.2,6X,F15.8)') trim(state_label(state)), &
                                state_charge(state), state_eigenvalue(state)/u
        end if
      case (2)
        if (qn%m == M_ZERO) then
          if (qn%s == -M_HALF) then
            nspin_label = "-1/2"
          else
            nspin_label = "+1/2"
          end if
          !write(message(j),'(7X,A,4X,A,5X,F5.2,4X,F12.5)') trim(state_label(state)), nspin_label, &
          write(message(j),'(7X,A,4X,A,5X,F5.2,4X,F15.8)') trim(state_label(state)), nspin_label, &
                                                           state_charge(state), state_eigenvalue(state)/u
        else
          if (qn%sg == M_HALF) then
            nspin_label = 'up'
          elseif (qn%sg == -M_HALF) then
            nspin_label = 'dn'
          else
            nspin_label = '--'
          end if
          !write(message(j),'(5X,A,4X,A,5X,F5.2,4X,F12.5)')  trim(state_label(state)), nspin_label, &
          write(message(j),'(5X,A,4X,A,5X,F5.2,4X,F15.8)')  trim(state_label(state)), nspin_label, &
                                                            state_charge(state), state_eigenvalue(state)/u
        end if
      end select
    end do

    if (present(unit)) then
      call write_info(j,unit=unit)
    else
      if (present(verbose_limit)) then
        call write_info(j,verbose_limit)
      else
        call write_info(j)
      end if
    end if

    call pop_sub()
  end subroutine states_batch_output_eigenvalues

    ! KL 05/02/2017 output the shell structure introduced in Flores&Keller PRA about internal symmetry
subroutine states_batch_output_shellstructure(batch, nspin, m, dir)
    !-----------------------------------------------------------------------!
    ! Writes the shell-structure to the "dir/shell" file in a format        !
    ! suitable for plotting.                                                !
    !-----------------------------------------------------------------------!
    type(states_batch_t), intent(in) :: batch
    integer,              intent(in) :: nspin
    type(mesh_t),         intent(in) :: m
    character(len=*),     intent(in) :: dir
    type(state_t), pointer :: state


    integer  :: i, r, p, unit
    real(R8) :: u, ne
    character(len=30) :: filename
    !real(R8), allocatable :: rho(:,:), rho_grad(:,:), rho_lapl(:,:), tau(:,:)
    real(R8), allocatable :: rho(:,:), ewd(:,:), ebar(:,:), shell(:, :)

    call push_sub("states_batch_output_shellstructure")

    allocate(rho(m%np, nspin), ewd(m%np, nspin))
    allocate(ebar(m%np, nspin), shell(m%np, nspin))
    rho      = states_batch_density     (batch, nspin, m)

    ewd      = M_ZERO
    ebar     = M_ZERO
    ne       = M_ZERO

    do i = 1, batch%n_states
            ewd = ewd +  batch%states(i)%ptr%ev * state_density(nspin, batch%states(i)%ptr)
            ebar= ebar + batch%states(i)%ptr%ev * batch%states(i)%ptr%occ
            ne  = ne + batch%states(i)%ptr%occ
    end do

    ewd = ewd/rho
    ebar = ebar/ne

    shell = ewd - ebar


    u = units_out%length%factor

    ! output the shell structure related quantities
    do p = 1, nspin
      if (nspin == 1) then
        filename = trim(dir)//"/shell"
      else
        if (p == 1) then
          filename = trim(dir)//"/shell_dn"
        elseif (p == 2) then
          filename = trim(dir)//"/shell_up"
        end if
      end if
      call io_open(unit, file=trim(filename))

      write(unit,'("#")')
      write(unit,'("# shell structure.")')
      write(unit,'("# Length units: ",A)') trim(units_out%length%name)
      write(unit,'("#")')

      write(unit,'("# ",71("-"))')
      write(unit,'("# |",7X,"r",7X,"|",6X,"ewd(r)",7X,"|",4X,"ebar(r)",4X,"|",4X,"shell(r)",4X,"|")')
      write(unit,'("# ",71("-"))')
      do i = 1, m%np
        write(unit,'(3X,ES14.8E2,3X,ES15.8E2,3X,ES15.8E2,3X,ES15.8E2)') &
             m%r(i)/u, ewd(i, p), ebar(i, p), shell(i, p)
      end do

      close(unit)
    end do

    !Free memory
    deallocate(rho, ewd, ebar, shell)

    call pop_sub()
  end subroutine states_batch_output_shellstructure



    ! KL 07/10/2017 added another file to output the exact Pauli potential
    ! KL 07/12/2017 added variable reduced density gradient q
    ! KL 07/15/2017 added enhancement factor for kinetic energy density
    ! KL 07/17/2017 added enhancement factor for pauli potential
  subroutine states_batch_output_density(batch, nspin, m, dir)
    !-----------------------------------------------------------------------!
    ! Writes the electronic density to the "dir/density" file in a format   !
    ! suitable for plotting.                                                !
    !-----------------------------------------------------------------------!
    type(states_batch_t), intent(in) :: batch
    integer,              intent(in) :: nspin
    type(mesh_t),         intent(in) :: m
    character(len=*),     intent(in) :: dir

    integer  :: i, p, ikedf, unit
    real(R8) :: u
    real(R8) :: c_TF
    real(R8) :: kappa, ei, eN, ifrac! kappa= 1/2(3 pi^2)^(-1/3), eNi=e_N - e_i, ifrac= |\phi(r)|^2/n(r)
    character(len=30) :: filename
    real(R8), allocatable :: rho(:,:), rho_grad(:,:), rho_lapl(:,:), rho_grad2(:,:), tau(:,:)
    real(R8), allocatable :: vtheta(:,:), vth2(:,:)

    real(R8), allocatable :: rdgs(:,:), rdgp(:,:), rdgq(:,:), ttheta(:,:), tvw(:,:)

    call push_sub("states_batch_output_density")

    !Get the density and its derivatives
    allocate(rho(m%np, nspin), rho_grad(m%np, nspin), rho_grad2(m%np, nspin))
    allocate(rho_lapl(m%np, nspin), tau(m%np, nspin))

    allocate(vtheta(m%np, nspin), vth2(m%np, nspin))

    ! define variables for computing the vtheta
    allocate(rdgs(m%np, nspin), rdgp(m%np, nspin), rdgq(m%np, nspin)) !reduced density gradients, s, p and q.
    allocate(tvw(m%np, nspin), ttheta(m%np, nspin))

    rho      = states_batch_density     (batch, nspin, m)
    rho_grad = states_batch_density_grad(batch, nspin, m)
    rho_lapl = states_batch_density_lapl(batch, nspin, m)
    tau      = states_batch_tau         (batch, nspin, m)

    do p = 1, nspin
        rho_grad2(:,p) = mesh_derivative(m, rho_grad(:,p))
    end do

    kappa = 0.5*(3.0*M_PI**2)**(-1.0/3.0)
    c_TF = 3.0/10.0*(3.0*M_PI**2)**(2.0/3.0)
    ! things to do.....
    vth2=0.0
    do i= 1, batch%n_states
        eN = batch%states(batch%n_states)%ptr%ev
        ei = batch%states(i)%ptr%ev
        !vth2 = vth2 - ei * state_density(nspin, batch%states(i)%ptr)*batch%states(i)%ptr%occ
        vth2 = vth2 - ei * state_density(nspin, batch%states(i)%ptr) * batch%states(i)%ptr%occ
        !write(*,*) i, ei, eN, eN-ei, nspin
    end do

    do p = 1, nspin
    if (p == 1) then
        do i = 1, m%np
        if (rho(i, p) > 10.0**(-16)) then
            rdgs(i,p) = kappa*rho_grad(i,p)*rho(i,p)**(-4.0/3.0)
            rdgp(i,p) = kappa**2*rho_lapl(i,p)*rho(i,p)**(-5.0/3.0)
            rdgq(i,p) = kappa**4*rho_grad(i,p)**2*rho_grad2(i,p)*rho(i,p)**(-13.0/3.0)
            !tvw(i,p)  = state_density_lapl(state, nspin, m)
            tvw(i,p)  = rho_grad(i,p)**2/8.0/rho(i,p)
            ttheta(i,p)=0.5*tau(i,p)-tvw(i,p)
            vtheta(i,p)=eN + ttheta(i,p)/rho(i,p) + vth2(i,p)/rho(i,p)
        else
            rdgs(i,p) = 0.0
            rdgp(i,p) = 0.0
            rdgq(i,p) = 0.0
            tvw(i,p) = 0.0
            ttheta(i,p)= 0.0
            vtheta(i,p)=0.0
        end if
        end do
    else
        write(*,*) ""
        write(*,*) "-----------------------------------------"
        write(*,*) "Implement spin-polarized case: to be done"
        write(*,*) "-----------------------------------------"
        write(*,*) ""
    end if
    end do





    u = units_out%length%factor

    ! print enhancement factor F[n]
    do p = 1, nspin
      if (nspin == 1) then
        filename = trim(dir)//"/enhancement"
      else
        if (p == 1) then
          filename = trim(dir)//"/enhancement_dn"
        elseif (p == 2) then
          filename = trim(dir)//"/enhancement_up"
        end if
      end if
      call io_open(unit, file=trim(filename))

      write(unit,'("#")')
      write(unit,'("#")')
      write(unit,'("# enhancement factor.")')
      write(unit,'("# Length units: ",A)') trim(units_out%length%name)
      write(unit,'("#")')

      write(unit,'("# ",41("-"))')
      write(unit,'("# |",7X,"r",7X,"|",6X,"F(r)",6X, "|", 6X, "Q(r)", 6X, "|", 6X, "s(r)", 6X, "p(r)", 6X, "|")')
      write(unit,'("# ",41("-"))')
      do i = 1, m%np
         write(unit,'(3X,ES14.8E2,3X,ES15.8E2,3X,ES15.8E2, 3X, ES15.8E2, 3X, ES15.8E2 )') &
             m%r(i)/u, ttheta(i,p)/rho(i,p)**(5.0/3.0)/c_TF, vtheta(i,p)/rho(i,p)**(2.0/3.0)/c_TF, rdgs(i,p),rdgp(i,p);
      end do

      close(unit)
    end do

    ! print factor t_theta =tau_theta / rho
    do p = 1, nspin
      if (nspin == 1) then
        filename = trim(dir)//"/t_theta"
      else
        if (p == 1) then
          filename = trim(dir)//"/t_theta_dn"
        elseif (p == 2) then
          filename = trim(dir)//"/t_theta_up"
        end if
      end if
      call io_open(unit, file=trim(filename))

      write(unit,'("#")')
      write(unit,'("#")')
      write(unit,'("# enhancement factor.")')
      write(unit,'("# Length units: ",A)') trim(units_out%length%name)
      write(unit,'("#")')

      write(unit,'("# ",71("-"))')
      write(unit,'("# |",7X,"r",7X,"|",5X,"ttheta(r)",5X, "|", 6X, "s2(r)", 6X, "|", 6X, "p(r)", 6X, "|", 6X, "rho(r)", 6X)')
      write(unit,'("# ",71("-"))')
      do i = 1, m%np
         write(unit,'(3X,ES14.8E2,3X,ES15.8E2,3X,ES15.8E2, 3X, ES15.8E2, 3X, ES15.8E2)') &
             m%r(i)/u, ttheta(i,p)/rho(i,p), rdgs(i,p)**2, rdgp(i,p), rho(i,p)*u;
      end do

      close(unit)
    end do


    ! print vtheta, s, p, q, rho
    do p = 1, nspin
      if (nspin == 1) then
        filename = trim(dir)//"/vtheta"
      else
        if (p == 1) then
          filename = trim(dir)//"/vtheta_dn"
        elseif (p == 2) then
          filename = trim(dir)//"/vtheta_up"
        end if
      end if
      call io_open(unit, file=trim(filename))

      write(unit,'("#")')
      write(unit,'("# Pauli potential.")')
      write(unit,'("# Length units: ",A)') trim(units_out%length%name)
      write(unit,'("#")')

      write(unit,'("# ",101("-"))')
      write(unit,'("# |",7X,"r",7X,"|",4X,"vtheta(r)",4X,"|",6X,"s^2(r)",6X,"|",6X,"p(r)",6X,"|", 6X, "q(r)",6X, "|", 5X, "rho(r)",&
      5X)')
      write(unit,'("# ",101("-"))')
      do i = 1, m%np
         write(unit,'(3X,ES14.8E2,3X,ES15.8E2,3X,ES15.8E2,3X,ES15.8E2,3X, ES15.8E2, 3X, ES15.8E2, 3X)') &
             m%r(i)/u, vtheta(i, p)/units_out%energy%factor, rdgs(i, p)**2, rdgp(i, p), rdgq(i,p), rho(i,p)*u
      end do

      !write(unit,'("# ",71("-"))')
      !write(unit,'("# |",7X,"1:r",7X,"|",6X,"2:vtheta(r)",7X,"|",4X,"3:s^2(r)",4X,"|",4X,"4:p(r)",4X,"|", 4X, "5:t_theta(r)", 4X, &
      !"|", 4X,  "6:vth2(r)", 4X, "|", 4X, "7:t_vw(r)", 4X, "|", 4X, "8:rho(r)", 4X)')
      !write(unit,'("# ",71("-"))')
      !do i = 1, m%np
      !  write(unit,'(3X,ES14.8E2,3X,ES15.8E2,3X,ES15.8E2,3X,ES15.8E2,3X, &
      !  ES15.8E2,3X,ES15.8E2,3X,ES15.8E2,3X,ES15.8E2,3X)') &
      !      m%r(i)/u, vtheta(i, p)/units_out%energy%factor, rdgs(i, p)**2, rdgp(i, p), ttheta(i,p), vth2(i,p), tvw(i,p), rho(i,p)
      !      m%r(i)/u, vtheta(i, p), rdgs(i, p)**2, rdgp(i, p), ttheta(i,p), vth2(i,p), tvw(i,p), rho(i,p)
      !end do

      close(unit)
    end do


    !Density, gradient, and laplacian
    do p = 1, nspin
      if (nspin == 1) then
        filename = trim(dir)//"/density"
      else
        if (p == 1) then
          filename = trim(dir)//"/density_dn"
        elseif (p == 2) then
          filename = trim(dir)//"/density_up"
        end if
      end if
      call io_open(unit, file=trim(filename))

      write(unit,'("#")')
      write(unit,'("# Radial density.")')
      write(unit,'("# Length units: ",A)') trim(units_out%length%name)
      write(unit,'("#")')

      write(unit,'("# ",71("-"))')
      write(unit,'("# |",7X,"r",7X,"|",6X,"n(r)",7X,"|",4X,"grad_n(r)",4X,"|",4X,"lapl_n(r)",4X,"|")')
      write(unit,'("# ",71("-"))')
      do i = 1, m%np
        write(unit,'(3X,ES14.8E2,3X,ES15.8E2,3X,ES15.8E2,3X,ES15.8E2)') &
             m%r(i)/u, rho(i, p)*u, rho_grad(i, p)*u**2, rho_lapl(i, p)*u**3
      end do

      close(unit)
    end do

    !Print kinetic energy density
    do p = 1, nspin
      if (nspin == 1) then
        filename = trim(dir)//"/tau"
      else
        if (p == 1) then
          filename = trim(dir)//"/tau_dn"
        elseif (p == 2) then
          filename = trim(dir)//"/tau_up"
        end if
      end if
      call io_open(unit, file=trim(filename))

      write(unit,'("#")')
      write(unit,'("# Radial kinetic energy density.")')
      write(unit,'("# Length units: ",A)') trim(units_out%length%name)
      write(unit,'("# Energy units: ",A)') trim(units_out%energy%name)
      write(unit,'("#")')

      write(unit,'("# ",35("-"))')
      write(unit,'("# |",7X,"r",7X,"|",5X,"tau(r)",6X,"|")')
      write(unit,'("# ",35("-"))')
      do i = 1, m%np
        write(unit,'(3X,ES14.8E2,3X,ES15.8E2)') m%r(i)/u, &
                                              tau(i, p)*u/units_out%energy%factor
      end do

      close(unit)
    end do

    !Free memory
    deallocate(rho, rho_grad, rho_lapl, tau)
    deallocate(tvw, ttheta, vtheta, rdgs, rdgp, vth2)


    call pop_sub()
  end subroutine states_batch_output_density

  subroutine states_batch_ps_io_set(batch, m, rc)
    !-----------------------------------------------------------------------!
    ! Pass the information about the wavefunctions to the ps_io module.     !
    !-----------------------------------------------------------------------!
    type(states_batch_t), intent(in) :: batch
    type(mesh_t),         intent(in) :: m
    real(R8),             intent(in) :: rc(batch%n_states)

    integer :: i
    character(len=10) :: label
    type(qn_t) :: qn
    integer, allocatable :: n(:), l(:)
    real(R8), allocatable :: j(:), occ(:), ev(:), wfs(:,:), rho_val(:,:)
    type(state_t), pointer :: state

    call push_sub("states_batch_ps_io_set")

    allocate(n(batch%n_states), l(batch%n_states), j(batch%n_states))
    allocate(occ(batch%n_states), ev(batch%n_states))
    allocate(wfs(m%np, batch%n_states))
    allocate(rho_val(m%np, 1))

    rho_val = states_batch_density(batch, 1, m)
    do i = 1, batch%n_states
      state => batch%states(i)%ptr
      qn = state_qn(state)
      label = state_label(state)
      read(label,'(I1)') n(i)
      l(i) = qn%l
      j(i) = qn%j
      occ(i) = state_charge(state)
      ev(i) = state_eigenvalue(state)
      wfs(:, i) = state%wf(:,1)
    end do

    call ps_io_set_wfs(m%np, batch%n_states, n, l, j, occ, ev, rc, wfs, &
                       sum(rho_val, dim=2))

    deallocate(n, l, j, occ, ev, wfs, rho_val)

    call pop_sub()
  end subroutine states_batch_ps_io_set

  subroutine pauli_potential(ikedf, m, nspin, states, vth, tth)
  ! generic implementation, 2018-01-27 23:03

    integer, intent(in)      :: ikedf
    type(mesh_t), intent(in) :: m
    integer, intent(in)      :: nspin
    type(states_batch_t), intent(in) :: states
    real(R8), intent(inout)  :: vth(m%np, nspin), tth(m%np, nspin) ! pauli potential

    real(R8) :: tmparray(m%np)
    !type(mesh_t)         :: mesh

    real(R8), allocatable :: rho(:,:), rho_grad(:,:), rho_lapl(:,:), tau(:,:)

    real(R8), allocatable :: s(:), p(:), q(:), s2(:)

    real(R8), allocatable :: ds2dn(:), ds2dgradn(:)
    real(R8), allocatable :: idmatrix(:)

    real(R8), allocatable :: rho_SI(:), rho_grad2(:) !SI=spin-independent
    real(R8), allocatable :: rho_grad_SI(:), rho_lapl_SI(:)


    real(R8), allocatable :: ftheta(:), fthetap(:), fthetapp(:)

    integer  :: i, j
    real(R8) :: u
    real(R8) :: c_TF, kappa


    real(R8) :: LKTa
    real(R8) :: modLKTa, modLKTb

    real(R8) :: VT84Falpha, VT84Fmu

    real(R8) ::  APBEKkappa, APBEKmu
    real(R8) ::  WPBEKkappa, WPBEKmu, scero,acero !VARIABLES DE HECTOR WPBEK
    real(R8) ::  WPBEKmup, acerop !VARIABLES DE HECTOR WPBEK-PAULI-SD


    real(R8) :: rhothreshold
    real(R8) :: sthreshold

    call push_sub("pauli_potential")

    allocate(rho(m%np, nspin), rho_grad(m%np, nspin), rho_lapl(m%np, nspin), tau(m%np, nspin))

    allocate(rho_SI(m%np), rho_grad2(m%np), rho_grad_SI(m%np), rho_lapl_SI(m%np))

    allocate(s(m%np),s2(m%np), p(m%np), q(m%np)) !reduced density gradients, s, p and q.
    allocate(ds2dn(m%np),ds2dgradn(m%np), idmatrix(m%np)) !ds/dn, ds/d(\grad n)

    allocate(ftheta(m%np), fthetap(m%np), fthetapp(m%np)) !enhancement factor, derivative and 2nd order derivative


    kappa = 0.5*(3.0*M_PI**2)**(-1.0/3.0)
    c_TF = 3.0/10.0*(3.0*M_PI**2)**(2.0/3.0)

    rho      = states_batch_density     (states, nspin, m)
    rho_grad = states_batch_density_grad(states, nspin, m)
    rho_lapl = states_batch_density_lapl(states, nspin, m)
    tau      = states_batch_tau         (states, nspin, m)

    rho_SI = rho(:,1) !+rho(:,2)
    rho_grad_SI=rho_grad(:,1)
    rho_lapl_SI=rho_lapl(:,1)

    rho_grad2 = mesh_derivative(m, rho_grad_SI)

    !write(*, *) rho_grad_SI
    !where (rho_SI < 1.0e-30_r8)
    !  rho_SI=1.0e-30_r8
    !endwhere
    vth=0.0
    tth=0.0
    s=0.0
    s2=0.0
    p=0.0
    q=0.0
    ds2dn=0.0
    ds2dgradn=0.0

    idmatrix=1.0

    !rhothreshold=5.e-10_r8
    !rhothreshold=1.e-09_r8
    rhothreshold=1.e-08_r8
    !rhothreshold=1.e-07_r8
    !rhothreshold=1.e-04_r8
    !rhothreshold=1.e-05_r8
    sthreshold  =1.e-10_r8
    where (rho_SI > rhothreshold)
      s  = abs(kappa*rho_grad_SI*rho_SI**(-4.0/3.0))
      s2 = (kappa*rho_grad_SI*rho_SI**(-4.0/3.0))**2
      p  = kappa**2*rho_lapl_SI*rho_SI**(-5.0/3.0)
      q  = kappa**4*rho_grad_SI**2*rho_grad2*rho_SI**(-13.0/3.0)
      ds2dn = -8.0/3.0*s2/rho_SI
      ds2dgradn = 2*kappa**2*rho_grad_SI*rho_SI**(-8.0/3.0)
    endwhere


    ! write(*,*) "max abs(tmparray)=", maxval(abs(tmparray))
    ! write(*,*) "min abs(tmparray)=", minval(abs(tmparray))
    ! write(*,*) "max \grad s2 = ", maxval(mesh_derivative(m,s2))
    ! write(*,*) "max ******** = ", maxval(2*kappa**2*rho_grad_SI &
    ! *rho_grad2/rho_SI**(8.0/3.0)+8.0/3.0*s2*rho_grad_SI/rho_SI)

    ! tmparray=mesh_derivative(m,s2)
    ! do i=1,m%np
    !   write(*,*) "r=", m%r(i), "rho=", rho_SI(i), "\grad s2 = ", tmparray(i), &
    !       "******** = ", 2*kappa**2*rho_grad_SI(i) &
    !       *rho_grad2(i)/rho_SI(i)**(8.0/3.0)-8.0/3.0*s2(i)*rho_grad_SI(i)/rho_SI(i)
    ! end do





    !write(*,*) size(s), size(s2), size(p), size(q)
    !write(*,*) size(vth), size(tth)
    !write(*,*) size(vth), size(tth)
    !write(*,*) size(vth), size(tth)

    !    where (s > 1.e10)
    !      s  = 1.e10
    !      s2 = s**2
    !    endwhere
    !
    !    where (p > 1.e10)
    !      p  = 1.e10
    !    endwhere
    !
    !    where (q > 1.e10)
    !      q  = 1.e10
    !    endwhere

    ! KL: 2018-05-21 write s, p and q for debugging
    !write(*,*)
    !write(*,*) "smax=",maxval(s), " smin=", minval(s)
    !write(*,*) "pmax=",maxval(p), " pmin=", minval(p)
    !write(*,*) "qmax=",maxval(q), " qmin=", minval(q)


    !ftheta=0.0
    !fthetap=0.0
    !fthetapp=0.0

    ! write(*,100) "s=", s
    ! write(*,100) "p=", p
    ! write(*,100) "q=", q

    !write(*,*) "iKEDF=", ikedf
    select case (ikedf)
    case (1) !LKT
      !LKTa=1.2
      call oct_parse_float('LKTa', 1.3_R8, LKTa)
      !write(*,*) "Using LKT, LKTa=", LKTa

      ! do i=1,m%np
      !   ftheta(i)   = M_ONE/cosh(LKTa*s(i))
      !   if (rho_SI(i) > rhothreshold) then
      !     if (s(i) < sthreshold ) then
      !       fthetap(i) = -LKTa**2/2.0 +(5.*LKTa**4*s2(i))/12.
      !       fthetapp(i) = (5*LKTa**4)/12. - (61*LKTa**6*s2(i))/120.
      !     else
      !       fthetap(i)  = -(LKTa/cosh(LKTa*s(i))*tanh(LKTa*s(i)))/(2.*s(i))
      !       fthetapp(i) = LKTa/cosh(LKTa*s(i))**3*(LKTa*s(i)*(-3 + cosh(2*LKTa*s(i)))+sinh(2*LKTa*s(i)))/(8.*s(i)**3)
      !     end if
      !   else
      !     vth(i,1) = c_TF*rho_SI(i)**(2.0/3.0)*(5.0/3.0*ftheta(i) -(2.0/3.0*s2(i) +2*p(i))*fthetap(i) + 4.0*(4.0/3.0*s(i)**4-q(i))*fthetapp(i))
      !     vth(i,2) = vth(i,1)
      !     tth(i,1) = c_TF*rho_SI(i)**(2.0/3.0)*ftheta(i)
      !     tth(i,2) = tth(i,1)
      !   end if
      ! end do

      where (rho_SI > rhothreshold)
        ftheta   = M_ONE/cosh(LKTa*s)
        where (s < sthreshold)
          fthetap = -LKTa**2/2. + (5*LKTa**4*s2)/12.
          fthetapp = (5*LKTa**4)/12. - (61*LKTa**6*s2)/120.
        elsewhere
          fthetap  = -(LKTa/cosh(LKTa*s)*tanh(LKTa*s))/(2.*s)
          fthetapp = LKTa/cosh(LKTa*s)**3*(LKTa*s*(-3 + cosh(2*LKTa*s))+sinh(2*LKTa*s))/(8.*s**3)
          !where (s>1.e02_r8)
          !fthetapp = 0.0
          !fthetapp = 1./cosh(LKTa*s)
          !elsewhere
          !!endwhereere
        endwhere
        !vth(:,1) = c_TF*rho_SI**(2.0/3.0)*(5.0/3.0*ftheta -(2.0/3.0*s2 +2*p)*fthetap + 4.0*(4.0/3.0*s**4-q)*fthetapp)
        !tth(:,1) = c_TF*rho_SI**(2.0/3.0)*ftheta
        !  vth(:,2) = vth(:,1)
        ! tth(:,2) = tth(:,1)
      endwhere
      !write(*,*) "nspin=", nspin


    case (2) !VT84F
      VT84Falpha = 1.2965
      VT84Fmu = 2.778
      !write(*,*) "Using VT84F"

      where (rho_SI > rhothreshold)
        where (s < sthreshold)
          ftheta  =1 + s**2*(VT84Falpha - VT84Fmu) + s**4*(-VT84Falpha + VT84Falpha*VT84Fmu + VT84Fmu**2)
          fthetap = VT84Falpha - VT84Fmu + 2*s**2*(-VT84Falpha + VT84Falpha*VT84Fmu + VT84Fmu**2)
          fthetapp= 2*(-VT84Falpha + VT84Falpha*VT84Fmu + VT84Fmu**2) -  &
           3*s**2*(VT84Falpha**2 + VT84Falpha**2*VT84Fmu +   2*VT84Falpha*VT84Fmu**2 + 2*VT84Fmu**3)
        elsewhere
          ftheta   = (1 - s**(-2))/EXP(s**4*VT84Falpha) + s**(-2) - &
            (s**2*VT84Fmu)/(EXP(s**2*VT84Falpha)*(1 + s**2*VT84Fmu))
          fthetap = (-1 + EXP(-(s**4*VT84Falpha)))/s**4 + (2*VT84Falpha)/EXP(s**4*VT84Falpha) - &
        VT84Fmu/(EXP(s**2*VT84Falpha)*(1 + s**2*VT84Fmu)) + &
        s**2*((-2*VT84Falpha)/EXP(s**4*VT84Falpha) + &
           (VT84Fmu*(VT84Falpha + VT84Fmu + s**2*VT84Falpha*VT84Fmu))/ &
            (EXP(s**2*VT84Falpha)*(1 + s**2*VT84Fmu)**2))

         fthetapp = (2*(1 - EXP(-(s**4*VT84Falpha))))/s**6 +  &
           (2*(-1 + s**(-2))*VT84Falpha)/EXP(s**4*VT84Falpha) - &
           (4*VT84Falpha)/(EXP(s**4*VT84Falpha)*s**2) - &
           (4*(-1 + s**(-2))*s**4*VT84Falpha**2)/EXP(s**4*VT84Falpha) - &
           (2*s**2*VT84Fmu**3)/ (EXP(s**2*VT84Falpha)*(1 + s**2*VT84Fmu)**3) + &
           (2*VT84Fmu**2)/(EXP(s**2*VT84Falpha)*(1 + s**2*VT84Fmu)**2) - &
           (2*s**2*VT84Falpha*VT84Fmu**2)/(EXP(s**2*VT84Falpha)*(1 + s**2*VT84Fmu)**2) + &
           (2*VT84Falpha*VT84Fmu)/ (EXP(s**2*VT84Falpha)*(1 + s**2*VT84Fmu)) - &
           (s**2*VT84Falpha**2*VT84Fmu)/(EXP(s**2*VT84Falpha)*(1 + s**2*VT84Fmu))

        endwhere
        !vth(:,1) = c_TF*rho_SI**(2.0/3.0)*(5.0/3.0*ftheta -(2.0/3.0*s2 +2*p)*fthetap + 4.0*(4.0/3.0*s**4-q)*fthetapp)
        !vth(:,2) = vth(:,1)
        !tth(:,1) = c_TF*rho_SI**(2.0/3.0)*ftheta
        !tth(:,2) = tth(:,1)
      endwhere
!!!!FUNCIONALES H.FRANCISCO (WPBEK) add 06/15/2021 
      case (3) !WPBEK
      WPBEKkappa= 0.641
      WPBEKmu   = 0.23889
      scero     = 4.0
      acero     = 3.0
      where (rho_SI > rhothreshold)
!!! ENHACENMENT FACTOR WPBEK
        ftheta   = 1.0 + WPBEKkappa - (5.0*s**2)/3.0 - &
                    WPBEKkappa/(1.0 + ((WPBEKmu*s**2)/WPBEKkappa)) + &
                    (1.0/(1.0+(exp(-acero*(s-scero)))))*((5.0*s**2)/3.0)
!!!FIRST DERIVATE RESPECT TO s**2
        fthetap  = (-5.0/3.0) + WPBEKmu/((1.0 + ((WPBEKmu*s**2)/WPBEKkappa))**2) + &
                   (5.0/3.0)/(1.0+(exp(-acero*(s-scero)))) + &
                   (5.0/6.0)*((s*acero*exp(-acero*(s-scero)))/ &
                   ((1.0+(exp(-acero*(s-scero))))**2))
!!!SECOND DERIVATE RESPECT TO s**2
        fthetapp = -2.0*(WPBEKmu**2)/((1.0 + ((WPBEKmu*s**2)/WPBEKkappa))**3) + &
                   (5.0/4.0)*((acero*exp(-acero*(s-scero)))/ &
                   s*((1.0+(exp(-acero*(s-scero))))**2)) -   &
                   (5.0/12.0)*((acero*acero*exp(-acero*(s-scero)))/ &
                   ((1.0+(exp(-acero*(s-scero))))**2)) +   &
                   (5.0/6.0)*((acero*acero*exp(-2.0*acero*(s-scero)))/ &
                   ((1.0+(exp(-acero*(s-scero))))**3))
        endwhere
!!!!FINISH WPBEK
!!!! START WPBEK-PAULI
     case (4) !WPBEK-PAULI-SD (Al atom) add 06/15/2021 
      WPBEKmup   = (5.0/3.0) - 0.23889
      acerop     = 0.27610
      where (rho_SI > rhothreshold)
!!! ENHACENMENT FACTOR WPBEK 
        ftheta   =       acerop*((exp(-WPBEKmup*s*s))/((1.0+50.0*s**4)**2)) +  &
                   (1.0-acerop)*((exp(-WPBEKmup*s*s))/((1.0+0.5*s**4)**2))
!!!FIRST DERIVATE RESPECT TO s**2
        fthetap  =  acerop*(((-4.0*50.0*s*s*exp(-WPBEKmup*s*s))/((1.0+50.0*s**4)**3)) -  &
                         ((WPBEKmup*exp(-WPBEKmup*s*s))/((1.0+50.0*s**4)**2)))  +  &
                   (1.0-acerop)*(((-4.0*0.5*s*s*exp(-WPBEKmup*s*s))/((1.0+0.5*s**4)**3)) -  &
                         ((WPBEKmup*exp(-WPBEKmup*s*s))/((1.0+0.5*s**4)**2)))
!!!SECOND DERIVATE RESPECT TO s**2
        fthetapp = acerop*(((-4.0*50.0*exp(-WPBEKmup*s*s))/((1.0+50.0*s**4)**3)) +  &
                         ((WPBEKmup*WPBEKmup*exp(-WPBEKmup*s*s))/((1.0+50.0*s**4)**2))  +  &
                        ((8.0*50.0*s*s*WPBEKmup*exp(-WPBEKmup*s*s))/((1.0+50.0*s**4)**3)) + &
                        ((24*50.0*50.0**s*s*s*s*WPBEKmup*exp(-WPBEKmup*s*s))/((1.0+50.0*s**4)**4))) + &
                        (1.0-acerop)*(((-4.0*0.5*exp(-WPBEKmup*s*s))/((1.0+0.5*s**4)**3)) +  &
                         ((WPBEKmup*WPBEKmup*exp(-WPBEKmup*s*s))/((1.0+0.5*s**4)**2))  +  &
                        ((8.0*0.5*s*s*WPBEKmup*exp(-WPBEKmup*s*s))/((1.0+0.5*s**4)**3)) + &
                        ((24*0.5*0.5*s*s*s*s*WPBEKmup*exp(-WPBEKmup*s*s))/((1.0+0.5*s**4)**4)))
      endwhere
!!!!FINISH WPBEK-PAULI-SD
!!!
    case (5) !modLKT
      call oct_parse_float('modLKTa', 1.2_R8, modLKTa)
      call oct_parse_float('modLKTb', 1.0_R8, modLKTb)

      !write(*,*) "Using modLKT, modLKTa=", modLKTa, " and modLKTb=", modLKTb

      where (rho_SI > rhothreshold)
        !ftheta   = M_ONE/cosh(LKTa*s)
        ftheta  = (1 + modLKTb*s**2)/cosh(modLKTa*s)
        where (s < sthreshold)
          !fthetap = (-modLKTa**2/2. + modLKTb)*s**2 + ((5*modLKTa**4)/12. - modLKTa**2*modLKTb)*s**4
          fthetap =    -modLKTa**2/2. + modLKTb + ((5*modLKTa**4)/12. - modLKTa**2*modLKTb)*s**2 + &
            ((-61*modLKTa**6)/240. + (5*modLKTa**4*modLKTb)/8.)*s**4
          !fthetap = -LKTa**2/2. + (5*LKTa**4*s2)/12.
        elsewhere
          !fthetap = -(s*Sech(modLKTa*s)*(-2*modLKTb*s + modLKTa*(1 + modLKTb*s**2)*Tanh(modLKTa*s)))/2.
          !fthetap = -(s/cosh(modLKTa*s)*(-2*modLKTb*s + modLKTa*(1 + modLKTb*s**2)*tanh(modLKTa*s)))/2.
          !fthetap = -(s/cosh(modLKTa*s)*( modLKTa*tanh(modLKTa*s)))/2.
          fthetap = -(1/cosh(modLKTa*s)*(-2*modLKTb*s + modLKTa*(1 + modLKTb*s**2)*tanh(modLKTa*s)))/(2.*s)

          !fthetap  = -(LKTa/cosh(LKTa*s)*tanh(LKTa*s))/(2.*s)
        endwhere
      endwhere
      !write(*,*) "nspin=", nspin


    case default
      !write(*,*) "Using von Weizsacker"
      ftheta=0.0
      fthetap=0.0
      !fthetapp=0.0
    end select

    vth(:,1)=c_TF*(5.0/3.0*rho_SI**(2.0/3.0)*ftheta + rho_SI**(5.0/3.0)*fthetap*ds2dn &
- m%r**(-2.0)*mesh_derivative(m,m%r**(2.0)*rho_SI**(5.0/3.0)*fthetap*ds2dgradn))
    tth(:,1)= c_TF*rho_SI**(2.0/3.0)*ftheta

    !write(*,*)
    !write(*,*) "diff_max(nabla s^2 - *)", mesh_derivative(m,s2) - 2*kappa**2*rho_grad_SI &
    !    *rho_grad2/rho_SI**(8.0/3.0)+8.0/3.0*s2*rho_grad_SI/rho_SI
    !write(*,*) "diff_max(nabla s^2 - *)", maxval(mesh_derivative(m,s2) - 2*kappa**2*rho_grad_SI &
    !    *rho_grad2/rho_SI**(8.0/3.0)+8.0/3.0*s2*rho_grad_SI/rho_SI), &
    !    "diff_min()", minval(mesh_derivative(m,s2) - 2*kappa**2*rho_grad_SI &
    !        *rho_grad2/rho_SI**(8.0/3.0)+8.0/3.0*s2*rho_grad_SI/rho_SI)
    !write(*,*) "diff_max=",maxval(mesh_derivative(m,rho_SI**(5.0/3.0)*fthetap*ds2dgradn)- &
    !    2*rho_SI**(2.0/3.0)*(fthetap*(p-s2)+fthetapp*(-8.0/3.0*s2**2 + 4.0*q)))

    ! tmparray=c_TF*m%r**(-2.0)*mesh_derivative(m,m%r**(2.0)*rho_SI**(5.0/3.0)*fthetap*ds2dgradn)
    ! do i=1,m%np
    !   write(*,*) "r=", m%r(i), "rho=", rho_SI(i), "generic fd = ", tmparray(i), &
    !       "method 2 = ",c_TF*2*rho_SI(i)**(2.0/3.0)*(fthetap(i)*(p(i)-s2(i))+fthetapp(i)*(-8.0/3.0*s2(i)**2 + 4.0*q(i)))
    ! end do

    ! write out the vtheta for debugging
    !write(*,*)
    !write(*,*) "vtheta_max=",maxval(vth), " vtheta_min=", minval(vth)
    !write(*,*) "ttheta_max=",maxval(tth), " ttheta_min=", minval(tth)

    !vth = vth/2
    !write(*,*) vth(:,1)
    !write(*,100) "ftheta=", ftheta
    !write(*,100) "fthetap=", fthetap
    !write(*,100) "fthetapp=", fthetapp


    !      where (rho_SI < 1.e-10_r8)
    !        vth(:,1) = 0.0
    !        vth(:,2) = 0.0
    !      endwhere

    !write(*,100) "vth=", vth(:,1)
    100 format (A, 4(2X, E16.10))

    deallocate(rho, rho_grad, rho_lapl, tau)
    deallocate(rho_SI, rho_grad_SI, rho_grad2, rho_lapl_SI)
    deallocate(s,s2,p,q)
    deallocate(ftheta,fthetap,fthetapp)
    deallocate(ds2dn, ds2dgradn, idmatrix)

    call pop_sub()
  end subroutine pauli_potential


!  subroutine pauli_potential(ikedf, m, nspin, states, vth, tth)
!
!    integer, intent(in)      :: ikedf
!    type(mesh_t), intent(in) :: m
!    integer, intent(in)      :: nspin
!    type(states_batch_t), intent(in) :: states
!    real(R8), intent(inout)  :: vth(m%np, nspin), tth(m%np, nspin) ! pauli potential
!
!    !type(mesh_t)         :: mesh
!
!    real(R8), allocatable :: rho(:,:), rho_grad(:,:), rho_lapl(:,:), tau(:,:)
!
!    real(R8), allocatable :: s(:), p(:), q(:), s2(:)
!    real(R8), allocatable :: rho_SI(:), rho_grad2(:) !SI=spin-independent
!    real(R8), allocatable :: rho_grad_SI(:), rho_lapl_SI(:)
!
!
!    real(R8), allocatable :: ftheta(:), fthetap(:), fthetapp(:)
!
!    integer  :: i, j
!    real(R8) :: u
!    real(R8) :: c_TF, kappa
!
!
!    real(R8) :: LKTa
!    real(R8) :: VT84Falpha, VT84Fmu
!
!    real(R8) ::  APBEKkappa, APBEKmu
!
!    real(R8) :: rhothreshold
!    real(R8) :: sthreshold
!
!    call push_sub("pauli_potential")
!
!    allocate(rho(m%np, nspin), rho_grad(m%np, nspin), rho_lapl(m%np, nspin), tau(m%np, nspin))
!
!    allocate(rho_SI(m%np), rho_grad2(m%np), rho_grad_SI(m%np), rho_lapl_SI(m%np))
!
!    allocate(s(m%np),s2(m%np), p(m%np), q(m%np)) !reduced density gradients, s, p and q.
!
!    allocate(ftheta(m%np), fthetap(m%np), fthetapp(m%np)) !enhancement factor, derivative and 2nd order derivative
!
!
!    kappa = 0.5*(3.0*M_PI**2)**(-1.0/3.0)
!    c_TF = 3.0/10.0*(3.0*M_PI**2)**(2.0/3.0)
!
!    rho      = states_batch_density     (states, nspin, m)
!    rho_grad = states_batch_density_grad(states, nspin, m)
!    rho_lapl = states_batch_density_lapl(states, nspin, m)
!    tau      = states_batch_tau         (states, nspin, m)
!
!    rho_SI = rho(:,1) !+rho(:,2)
!    rho_grad_SI=rho_grad(:,1)
!    rho_lapl_SI=rho_lapl(:,1)
!
!    rho_grad2 = mesh_derivative(m, rho_grad_SI)
!
!    !write(*, *) rho_grad_SI
!    !where (rho_SI < 1.0e-30_r8)
!    !  rho_SI=1.0e-30_r8
!    !endwhere
!    vth=0.0
!	  tth=0.0
!    s=0.0
!    s2=0.0
!    p=0.0
!    q=0.0
!
!    !rhothreshold=5.e-10_r8
!    !rhothreshold=1.e-09_r8
!    rhothreshold=1.e-08_r8
!    !rhothreshold=1.e-07_r8
!    !rhothreshold=1.e-04_r8
!    !rhothreshold=1.e-05_r8
!    sthreshold  =1.e-10_r8
!    where (rho_SI > rhothreshold)
!      s  = abs(kappa*rho_grad_SI*rho_SI**(-4.0/3.0))
!      s2 = (kappa*rho_grad_SI*rho_SI**(-4.0/3.0))**2
!      p  = kappa**2*rho_lapl_SI*rho_SI**(-5.0/3.0)
!      q  = kappa**4*rho_grad_SI**2*rho_grad2*rho_SI**(-13.0/3.0)
!    endwhere
!
!    !write(*,*) size(s), size(s2), size(p), size(q)
!    !write(*,*) size(vth), size(tth)
!    !write(*,*) size(vth), size(tth)
!    !write(*,*) size(vth), size(tth)
!
!    !    where (s > 1.e10)
!    !      s  = 1.e10
!    !      s2 = s**2
!    !    endwhere
!    !
!    !    where (p > 1.e10)
!    !      p  = 1.e10
!    !    endwhere
!    !
!    !    where (q > 1.e10)
!    !      q  = 1.e10
!    !    endwhere
!
!    write(*,*)
!    write(*,*) "smax=",maxval(s), " smin=", minval(s)
!    write(*,*) "pmax=",maxval(p), " pmin=", minval(p)
!    write(*,*) "qmax=",maxval(q), " qmin=", minval(q)
!
!
!    !ftheta=0.0
!    !fthetap=0.0
!    !fthetapp=0.0
!
!    ! write(*,100) "s=", s
!    ! write(*,100) "p=", p
!    ! write(*,100) "q=", q
!
!    !write(*,*) "iKEDF=", ikedf
!    select case (ikedf)
!    case (1) !LKT
!      !LKTa=1.2
!      call oct_parse_float('LKTa', 1.2_R8, LKTa)
!      write(*,*) "Using LKT, LKTa=", LKTa
!
!      ! do i=1,m%np
!      !   ftheta(i)   = M_ONE/cosh(LKTa*s(i))
!      !   if (rho_SI(i) > rhothreshold) then
!      !     if (s(i) < sthreshold ) then
!      !       fthetap(i) = -LKTa**2/2.0 +(5.*LKTa**4*s2(i))/12.
!      !       fthetapp(i) = (5*LKTa**4)/12. - (61*LKTa**6*s2(i))/120.
!      !     else
!      !       fthetap(i)  = -(LKTa/cosh(LKTa*s(i))*tanh(LKTa*s(i)))/(2.*s(i))
!      !       fthetapp(i) = LKTa/cosh(LKTa*s(i))**3*(LKTa*s(i)*(-3 + cosh(2*LKTa*s(i)))+sinh(2*LKTa*s(i)))/(8.*s(i)**3)
!      !     end if
!      !   else
!      !     vth(i,1) = c_TF*rho_SI(i)**(2.0/3.0)*(5.0/3.0*ftheta(i) -(2.0/3.0*s2(i) +2*p(i))*fthetap(i) + 4.0*(4.0/3.0*s(i)**4-q(i))*fthetapp(i))
!      !     vth(i,2) = vth(i,1)
!      !     tth(i,1) = c_TF*rho_SI(i)**(2.0/3.0)*ftheta(i)
!      !     tth(i,2) = tth(i,1)
!      !   end if
!      ! end do
!
!      where (rho_SI > rhothreshold)
!        ftheta   = M_ONE/cosh(LKTa*s)
!        where (s < sthreshold)
!          fthetap  = -LKTa**2/2. + (5*LKTa**4*s2)/12.
!          fthetapp = (5*LKTa**4)/12. - (61*LKTa**6*s2)/120.
!        elsewhere
!          fthetap  = -(LKTa/cosh(LKTa*s)*tanh(LKTa*s))/(2.*s)
!          fthetapp = LKTa/cosh(LKTa*s)**3*(LKTa*s*(-3 + cosh(2*LKTa*s))+sinh(2*LKTa*s))/(8.*s**3)
!          !where (s>1.e02_r8)
!          !fthetapp = 0.0
!          !fthetapp = 1./cosh(LKTa*s)
!          !elsewhere
!          !!endwhereere
!        endwhere
!        vth(:,1) = c_TF*rho_SI**(2.0/3.0)*(5.0/3.0*ftheta -(2.0/3.0*s2 +2*p)*fthetap + 4.0*(4.0/3.0*s**4-q)*fthetapp)
!        tth(:,1) = c_TF*rho_SI**(2.0/3.0)*ftheta
!        !  vth(:,2) = vth(:,1)
!		    ! tth(:,2) = tth(:,1)
!      endwhere
!      !write(*,*) "nspin=", nspin
!
!
!    case (2) !VT84F
!      VT84Falpha = 1.2965
!      VT84Fmu = 2.778
!      write(*,*) "Using VT84F"
!
!      where (rho_SI > rhothreshold)
!        where (s < sthreshold)
!          ftheta  = 1 + s**2*(VT84Falpha - VT84Fmu)
!          fthetap =  VT84Falpha - VT84Fmu + 2*s**2*(-VT84Falpha + VT84Falpha*VT84Fmu + VT84Fmu**2)
!          fthetapp= 2*(-VT84Falpha + VT84Falpha*VT84Fmu + VT84Fmu**2) -  &
!            3*s**2*(VT84Falpha**2 + VT84Falpha**2*VT84Fmu +   2*VT84Falpha*VT84Fmu**2 + 2*VT84Fmu**3)
!        elsewhere
!          ftheta   = (1 - s**(-2))/EXP(s**4*VT84Falpha) + s**(-2) - &
!            (s**2*VT84Fmu)/(EXP(s**2*VT84Falpha)*(1 + s**2*VT84Fmu))
!          fthetap  = (-1 + EXP(-(s**4*VT84Falpha)))/s**4 + (2*VT84Falpha)/EXP(s**4*VT84Falpha) &
!            - VT84Fmu/(EXP(s**2*VT84Falpha)*(1 + s**2*VT84Fmu)) + &
!            s**2*((-2*VT84Falpha)/EXP(s**4*VT84Falpha) + &
!            (VT84Fmu*(VT84Falpha + VT84Fmu + s**2*VT84Falpha*VT84Fmu))/ &
!            (EXP(s**2*VT84Falpha)*(1 + s**2*VT84Fmu)**2))
!          !fthetapp =  (2 - 2/EXP(s**4*VT84Falpha))/s**6 - (2*VT84Falpha)/EXP(s**4*VT84Falpha) &
!          !  - (2*VT84Falpha)/(EXP(s**4*VT84Falpha)*s**2) + (4*s**4*VT84Falpha**2)/EXP(s**4*VT84Falpha) &
!          !  + (2*VT84Fmu*(VT84Falpha + VT84Fmu + s**2*VT84Falpha*VT84Fmu))&
!          !  / (EXP(s**2*VT84Falpha)*(1 + s**2*VT84Fmu)**2) - (s**2*(4*EXP(s**2*VT84Falpha)*VT84Falpha**2* (1 + s**2*VT84Fmu)**3 &
!          !  + EXP(s**4*VT84Falpha)*VT84Fmu* (2*VT84Fmu**2 + 2*VT84Falpha*VT84Fmu*(1 + s**2*VT84Fmu)  &
!          !  + (VT84Falpha + s**2*VT84Falpha*VT84Fmu)**2)))/ (EXP(s**2*(1 + s**2)*VT84Falpha)*(1 + s**2*VT84Fmu)**3)
!          fthetapp = (2*(1 - EXP(-(s**4*VT84Falpha))))/s**6 +  &
!            (2*(-1 + s**(-2))*VT84Falpha)/EXP(s**4*VT84Falpha) - &
!            (4*VT84Falpha)/(EXP(s**4*VT84Falpha)*s**2) - &
!            (4*(-1 + s**(-2))*s**4*VT84Falpha**2)/EXP(s**4*VT84Falpha) - &
!            (2*s**2*VT84Fmu**3)/ (EXP(s**2*VT84Falpha)*(1 + s**2*VT84Fmu)**3) + &
!            (2*VT84Fmu**2)/(EXP(s**2*VT84Falpha)*(1 + s**2*VT84Fmu)**2) - &
!            (2*s**2*VT84Falpha*VT84Fmu**2)/(EXP(s**2*VT84Falpha)*(1 + s**2*VT84Fmu)**2) + &
!            (2*VT84Falpha*VT84Fmu)/ (EXP(s**2*VT84Falpha)*(1 + s**2*VT84Fmu)) - &
!            (s**2*VT84Falpha**2*VT84Fmu)/(EXP(s**2*VT84Falpha)*(1 + s**2*VT84Fmu))
!        endwhere
!        vth(:,1) = c_TF*rho_SI**(2.0/3.0)*(5.0/3.0*ftheta -(2.0/3.0*s2 +2*p)*fthetap + 4.0*(4.0/3.0*s**4-q)*fthetapp)
!        !vth(:,2) = vth(:,1)
!		    tth(:,1) = c_TF*rho_SI**(2.0/3.0)*ftheta
!		    !tth(:,2) = tth(:,1)
!      endwhere
!
!    case (3) !APBEK
!      APBEKkappa = 0.804
!      APBEKmu    = 0.23889
!      write(*,*) "Using APBEK"
!
!      where (rho_SI > rhothreshold)
!        ftheta   = 1 + APBEKkappa - (5*s**2)/3. - APBEKkappa/(1 + (APBEKkappa*s**2)/APBEKmu)
!        fthetap  = -1.6666666666666667 + APBEKkappa**2/(APBEKmu*(1 + (APBEKkappa*s**2)/APBEKmu)**2)
!        fthetapp = (-2*APBEKkappa**3)/(APBEKmu**2*(1 + (APBEKkappa*s**2)/APBEKmu)**3)
!        vth(:,1) = c_TF*rho_SI**(2.0/3.0)*(5.0/3.0*ftheta -(2.0/3.0*s2 +2*p)*fthetap + 4.0*(4.0/3.0*s**4-q)*fthetapp)
!        !vth(:,2) = vth(:,1)
!		    tth(:,1) = c_TF*rho_SI**(2.0/3.0)*ftheta
!		!tth(:,2) = tth(:,1)
!      endwhere
!
!    case (4) !TF
!      write(*,*) "Using Thomas-Fermi"
!
!      where (rho_SI > rhothreshold)
!        ftheta   = 1 - (5*s**2)/3.
!        fthetap  =-1.6666666666666667
!        fthetapp = 0.
!        vth(:,1) = c_TF*rho_SI**(2.0/3.0)*(5.0/3.0*ftheta -(2.0/3.0*s2 +2*p)*fthetap + 4.0*(4.0/3.0*s**4-q)*fthetapp)
!        !vth(:,2) = vth(:,1)
!		    tth(:,1) = c_TF*rho_SI**(2.0/3.0)*ftheta
!		    !tth(:,2) = tth(:,1)
!      endwhere
!
!
!    case default
!      write(*,*) "Using von Weizsacker"
!      tth=0.0
!      vth=0.0
!    end select
!
!
!
!    write(*,*)
!    write(*,*) "vtheta_max=",maxval(vth), " vtheta_min=", minval(vth)
!    write(*,*) "ttheta_max=",maxval(tth), " ttheta_min=", minval(tth)
!
!    !vth = vth/2
!    !write(*,*) vth(:,1)
!    !write(*,100) "ftheta=", ftheta
!    !write(*,100) "fthetap=", fthetap
!    !write(*,100) "fthetapp=", fthetapp
!
!
!    !      where (rho_SI < 1.e-10_r8)
!    !        vth(:,1) = 0.0
!    !        vth(:,2) = 0.0
!    !      endwhere
!
!    !write(*,100) "vth=", vth(:,1)
!    100 format (A, 4(2X, E16.10))
!
!    deallocate(rho, rho_grad, rho_lapl, tau)
!    deallocate(rho_SI, rho_grad_SI, rho_grad2, rho_lapl_SI)
!    deallocate(s,s2,p,q)
!    deallocate(ftheta,fthetap,fthetapp)
!    call pop_sub()
!  end subroutine pauli_potential


  subroutine output_pauli_potential(m, nspin, states)

    !integer, intent(in)      :: ikedf
    type(mesh_t), intent(in) :: m
    integer, intent(in)      :: nspin
    type(states_batch_t), intent(in) :: states
    real(R8), allocatable  :: vth(:, :), tth(:,:) ! pauli potential
    !real(R8)  :: vth(m%np, nspin) ! pauli potential
    integer :: postikedf
    integer :: i, unit
    real(R8) :: u

    call push_sub("output_pauli_potential")

    !write(*,*) "within output_pauli_potential"

    allocate(vth(m%np,nspin))
    allocate(tth(m%np,nspin))


    call oct_parse_int('postikedf', 0, postikedf)
    !write(*,*) "output pauli potential is called, iKEDF=", ikedf
    if ( postikedf /= 0 ) then
        call pauli_potential(postikedf, m, nspin, states, vth, tth)
        !write(*,*) "after call of pauli_potential"


        !unit=100
        call io_open(unit, file='of/postscf_vtheta')

        u = units_out%length%factor

        do i=1, m%np
        write(unit, 100) m%r(i)/u, vth(i,1), tth(i,1)
        !write(*,*) "vth(i)=",vth(i,1)
        end do

        close(unit)
        !write(*,*) "exiting output_pauli_potential"
    else
        write(*,*) "postikedf = ", postikedf
        write (*,*) "Won't output post-scf Pauli-potential"
    end if

    deallocate(vth)

    100 format (3(2X, E16.10))

    call pop_sub()
  end subroutine output_pauli_potential




end module states_batch_m
